<?php

namespace Modules\NsBulkImporter\Console\Commands;

use App\Models\User;
use Exception;
use Illuminate\Console\Command;
use Illuminate\Support\Facades\Auth;
use Modules\NsBulkImporter\Services\ImporterService;

class BulkImportCommand extends Command
{
    protected $signature = 'ns:bulkimport {path} {--config=} {--delimiter=,} {--email=null}';

    protected $description = 'Will parse and import a JSON config file.';

    /**
     *  @var ImporterService
     */
    public $importerService;

    public function __construct(
        ImporterService $importerService
    ) {
        parent::__construct();

        $this->importerService = $importerService;
    }

    public function handle()
    {
        if (is_file(base_path($this->argument('path')))) {
            $config = json_decode(file_get_contents(base_path($this->option('config'))), true);
            $file = base_path($this->argument('path'));
            $csv = array_map('str_getcsv', file($file));

            array_walk($csv, function (&$a, $index) use ($csv) {
                try {
                    $a = array_combine($csv[0], $a);
                } catch (Exception $exception) {
                    throw new Exception('Header And Row mismatch on line:'.$index);
                }
            });

            array_shift($csv);

            /**
             * mapping columns
             */
            $import = collect($csv)->map(function ($row, $index) use ($config) {
                $result = [];
                foreach ($row as $key => $value) {
                    $result[$config['columns'][$key]]['value'] = $value;
                }

                return $result;
            });

            /**
             * default values that will be used for each imports
             */
            $defaults = [
                'default_category'      =>      $config['default']['category'],
                'default_unit'          =>      $config['default']['unit'],
                'default_status'        =>      $config['default']['status'],
                'default_type'          =>      $config['default']['type'],
            ];

            /**
             * as all imported resources needs
             * to be assigned to a user, let's ask to which user
             * the import should be assigned to.
             */
            $email = $this->option('email');
            if (in_array($this->option('email'), [null, 'null'])) {
                $email = $this->ask(__('To which user the current import should be assigned to (email needed) ?'));
            }

            $user = User::where('email', $email)->firstOrFail();
            Auth::loginUsingId($user->id);

            /**
             * start importing categories
             */
            $this->info('Importing Categories...');

            $result = $this->importerService->createCategories($import, $defaults);
            $import = $result['data']['products'];
            $this->info(sprintf('%s categories imported/updated.', collect($import)
                ->filter(fn ($product) => isset($product['category']['id']))
                ->mapWithKeys(fn ($product) => [$product['category']['id'] => $product['category']['id']])
                ->count()));

            /**
             * Start importing units
             */
            $this->info('Importing Units...');
            $result = $this->importerService->createUnits($import, $defaults);
            $import = $result['data']['products'];
            $this->info(sprintf('%s units imported/updated.', collect($import)
                ->filter(fn ($product) => isset($product['unit']['id']))
                ->mapWithKeys(fn ($product) => [$product['unit']['id'] => $product['unit']['id']])
                ->count()));

            /**
             * start importing products
             */
            $this->info('Importing Products...');
            $result = $this->importerService->createProducts($import, $defaults);
            $finalResult = $result['data']['finalResult'];

            return $this->info(sprintf('Import completed : %s products imported/updated.', count($finalResult['success'])));
        }

        return $this->error(sprintf(__('Unable to located the provided file on the local filesystem : %s'), $this->argument('path')));
    }
}
