<?php

namespace Modules\NsBulkImporter\Services;

use App\Models\Customer;
use App\Models\CustomerGroup;
use App\Models\PaymentType;
use App\Models\Procurement;
use App\Models\Product;
use App\Models\ProductCategory;
use App\Models\Provider;
use App\Models\RewardSystem;
use App\Models\Unit;
use App\Services\BarcodeService;
use App\Services\CustomerService;
use App\Services\Options;
use App\Services\OrdersService;
use App\Services\ProcurementService;
use App\Services\ProductCategoryService;
use App\Services\ProductService;
use App\Services\ProviderService;
use Carbon\Carbon;
use Illuminate\Support\Str;
use Exception;
use Illuminate\Support\Facades\Auth;

class ImporterService
{
    public function __construct(
        protected ProductService $productService,
        protected ProductCategoryService $productCategoryService,
        protected CustomerService $customerService,
        protected ProviderService $providerService,
        protected Options $optionsService,
        protected ProcurementService $procurementService,
    ) {
        //
    }

    public function loadManifest( $requestedFile )
    {
        $file   =   base_path( 'modules/NsBulkImporter/Public/js/mix-manifest.json' );

        if ( is_file( $file ) ) {
            $content    =   file_get_contents( $file );
            $parsed     =   json_decode( $content, true );
            $fileName   =   pathinfo( $parsed[ '/' . $requestedFile ])[ 'basename' ];

            return asset( 'modules/nsbulkimporter/js/' . $fileName ) ?? null;
        }

        return null;
    }

    public function createCategories($products, $default)
    {
        $defaultCategory = ProductCategory::find($default['default_category'])->first();

        $products = collect($products)->map(function ($row) use ($defaultCategory) {
            if (isset($row['category']['value']) && ! empty($row['category']['value'])) {

                /**
                 * we'll fetch the category using the provided name
                 * if that exist, we use the provided identifier
                 */
                $category = ProductCategory::where('name', $row['category']['value'])->first();

                if (! $category instanceof ProductCategory) {
                    $result = $this->productCategoryService->create([
                        'name'          =>  $row['category']['value'],
                        'preview_url'   =>  $row['category_preview']['value'] ?? '',
                    ]);

                    $category = $result['data']['category'];
                }

                /**
                 * that identifier will be used while creating the products
                 */
                $row['category']['id'] = $category->id;
            } else {
                $row['category'] = [
                    'id'    =>  $defaultCategory->id,
                ];
            }

            return $row;
        })->toArray();

        return [
            'status'    =>  'success',
            'message'   =>  __('The categories has been created'),
            'data'      =>  compact('products'),
        ];
    }

    public function createUnits($products, $default)
    {
        $hasUnit = collect($products)
            ->filter(fn ($row) => isset($row['unit']) && ! empty($row['unit']))
            ->count() > 0;

        $defaultUnit = Unit::findOrFail($default['default_unit']);

        $products = collect($products)->map(function ($product) use ($defaultUnit) {
            if (isset($row['unit']['value']) && ! empty($row['unit']['value'])) {
                $unit = Unit::where('name', $row['unit']['value'])->first();

                if (! $unit instanceof Unit) {
                    $product['unit']['id'] = $defaultUnit->id;
                } else {
                    $product['unit'] = [
                        'id'    =>  $unit->id,
                    ];
                }
            } else {
                $product['unit'] = [
                    'id'    =>  $defaultUnit->id,
                ];
            }

            return $product;
        })->toArray();

        return [
            'status'    =>  'success',
            'message'   =>  __('The Units has been created.'),
            'data'      =>  compact('products'),
        ];
    }

    public function createProducts($products, $default)
    {
        $finalResult = [
            'success'   =>  [],
            'failures'  =>  [],
        ];

        $products = collect($products)->map(function ($row) use ($default, &$finalResult) {
            $unit = Unit::find($row['unit']['id'] ?? $default['default_unit']);
            $randomValue = date('Y').date('m').rand(10000000, 99999999);

            try {
                $product = Product::where('barcode', $row['barcode']['value'] ?? '')
                    ->orWhere('sku', $row['sku']['value'] ?? '')
                    ->first();

                $images = [];

                /**
                 * if the image is defined, we'll
                 * save it as an image gallery
                 */
                if (isset($row['preview']['value'])) {
                    $images[] = [
                        'featured'   =>  true,
                        'url'       =>  $row['preview']['value'],
                    ];
                }

                $data = [
                    'product_type'                  =>      'product',
                    'name'                          =>      $row['name']['value'],
                    'sku'                           =>      ($row['sku']['value'] ?? $randomValue),
                    'barcode'                       =>      ($row['barcode']['value'] ?? $randomValue),
                    'category_id'                   =>      ($row['category']['id'] ?? $default['default_category']),
                    'units'                         =>      [
                        'unit_group'                =>      $unit->group->id,
                        'selling_group'             =>      [
                            [
                                'unit_id'               =>      $unit->id,
                                'preview_url'           =>      ($row['unit_preview']['value'] ?? ''),
                                'sale_price_edit'       =>      ($row['sale_price']['value'] ?? 0),
                                'wholesale_price_edit'  =>      ($row['wholesale_price']['value'] ?? 0),
                            ],
                        ],
                    ],
                    'tax_group_id'                  =>      0,
                    'status'                        =>      $default['default_status'] ?? 'unavailable',
                    'type'                          =>      $default['default_type'] ?? 'dematerialized',
                    'stock_management'              =>      $row['stock_management']['value'] ?? 'disabled',
                    'tax_type'                      =>      'inclusive',
                    'barcode_type'                  =>      'code128',
                    'images'                        =>      $images,
                ];

                if ($product instanceof Product) {
                    $result = $this->productService->update($product, $data);
                    $row = array_merge($row, [
                        'id'        =>  $result['data']['product']['id'],
                        'name'      =>  $result['data']['product']['name'],
                        'barcode'   =>  $result['data']['product']['barcode'],
                    ]);

                    $finalResult['success'][] = $result;
                } else {
                    $result = $this->productService->create($data);
                    $row = array_merge($row, [
                        'id'        =>  $result['data']['product']['id'],
                        'name'      =>  $result['data']['product']['name'],
                        'barcode'   =>  $result['data']['product']['barcode'],
                    ]);

                    $finalResult['success'][] = $result;
                }
            } catch (Exception $exception) {
                $finalResult['failures'][] = $exception->getMessage();
            }

            return $row;
        });

        return [
            'status'    =>  'success',
            'message'   =>  __('The products has been created.'),
            'data'      =>  compact('finalResult', 'products'),
        ];
    }

    /**
     * Will create customer group using
     * the provided information
     *
     * @param  array  $entries
     * @param  array  $form
     * @return array $entries update
     */
    public function createCustomerGroup($customers, $form)
    {
        $defaultGroup = CustomerGroup::findOrFail($form['default_group']);
        $rewardSystem = RewardSystem::find($form['default_reward_system']);

        foreach ($customers as &$entry) {
            $result = $this->customerService->createGroup([
                'name'              =>  $entry['group_name']['value'] ?? ($defaultGroup->name),
                'reward_system_id'  =>  $entry['reward_system']['value'] ?? ($rewardSystem->id ?? null),
            ]);

            $group = $result['data']['group'];
            $entry['group_name']['id'] = $group->id;
        }

        return [
            'status'    =>  'success',
            'message'   =>  __('The customer group has been created.'),
            'data'      =>  compact('customers'),
        ];
    }

    public function createCustomerFromOrder( $entries, $form )
    {          
        $bag                    =       collect([]);  
        $defaultCustomer        =       Customer::find( $form[ 'customer_id' ] );
        $entries                =       collect( $entries )->map( function( $entry ) use ( $defaultCustomer, $bag, $form ) {
            $found  =   null;
            /**
             * on the current bag, we'll search for the
             * customer that has already been created to assign
             * them without any futher DB Request.
             */
            if ( isset( $entry[ 'customer_email' ] ) && $found === null ) {
                $found  =   $bag->where( 'email', $entry[ 'customer_email' ][ 'value' ] )->first();
            }

            if ( isset( $entry[ 'customer_name' ] ) && $found === null ) {
                $found  =   $bag->where( 'name', $entry[ 'customer_name' ][ 'value' ] )->first();
            }

            if ( isset( $entry[ 'customer_phone' ] ) && $found === null ) {
                $found  =   $bag->where( 'phone', $entry[ 'customer_phone' ][ 'value' ] )->first();
            }

            /**
             * if nothing has been found, then the record is not on the bag
             * we'll then either check if that exists as a customer or use
             * the default customer if not valid datas are provided.
             */
            if ( $found === null ) {
                $existingQuery   =   Customer::query();

                if ( isset( $entry[ 'customer_name' ] ) ) {
                    $existingQuery->orWhere( 'name', $entry[ 'customer_name' ][ 'value' ] );
                }

                if ( isset( $entry[ 'customer_email' ] ) ) {
                    $existingQuery->orWhere( 'email', $entry[ 'customer_email' ][ 'value' ] );
                }

                if ( isset( $entry[ 'customer_phone' ] ) ) {
                    $existingQuery->orWhere( 'phone', $entry[ 'customer_phone' ][ 'value' ] );
                }

                $customer   =   $existingQuery->first();

                /**
                 * if there is no match for
                 * the provided customer, we'll create a new one.
                 */
                if ( ! $customer instanceof Customer ) {
                    /**
                     * if the name is not provided, we'll use
                     * the default customer instead regardless of the provided email.
                     */
                    if ( ! empty( $entry[ 'customer_name' ][ 'value' ] ) ) {
                        $domain     = parse_url( url('/') );
                        $result     =   $this->customerService->create([
                            'name'      =>  $entry[ 'customer_name' ][ 'value' ],
                            'email'     =>  $entry[ 'customer_email' ][ 'value' ] ?? 'customer-' . strtolower( Str::random(5) ) . '@' . ( $domain[ 'host' ] ?? 'nexopos.com' ),
                            'phone'     =>  $entry[ 'customer_phone' ][ 'value' ] ?? '',
                            'group_id'  =>  $form[ 'customer_group_id'] ?? 0,
                            'address'   =>  [
                                'shipping'  =>  [],
                                'billing'   =>  [],
                            ]
                        ]);

                        $entry[ 'customer' ]    =   $result[ 'data' ][ 'customer' ];
                    } else {
                        $entry[ 'customer' ]    =   $defaultCustomer;
                    }                    
                } else {
                    $entry[ 'customer' ]    =   $customer;
                }

                /**
                 * let's put a reference of this customer
                 * on the bag for further queries.
                 */
                $bag->push( $entry[ 'customer' ]  );

            } else {
                $entry[ 'customer' ]    =   $found;
            }

            return $entry;
        });

        return [
            'status'    =>  'success',
            'message'   =>  __( 'The customers were imported successfully.' ),
            'data'      =>  compact( 'entries' )
        ];
    }

    public function importProductsFromOrders( $entries, $form )
    {
        $result = [
            'success'   =>  [],
            'failures'  =>  [],
        ];

        $entries = collect($entries)->map(function ($entry) use ($form,&$result) {
            $unit           =   isset( $entry[ 'product_unit' ][ 'value' ] ) ? Unit::where( 'name', $entry['product_unit']['value'] )->first() : Unit::find($form['unit_id']);
            $foundCategory  =   ProductCategory::where( 'name', $entry[ 'product_category' ][ 'value' ])->orWhere( 'id', $entry[ 'product_category' ][ 'value' ])->first();
            $category       =   $foundCategory instanceof ProductCategory ? $foundCategory : ProductCategory::find( $form[ 'category_id' ] );
            $randomValue    =   date('Y').date('m').rand(10000000, 99999999);

            try {
                $product = Product::where('barcode', $entry['product_barcode']['value'] ?? '')
                    ->orWhere('sku', $entry['product_sku']['value'] ?? '')
                    ->first();

                $images = [];

                /**
                 * if the image is defined, we'll
                 * save it as an image gallery
                 */
                if (isset($entry['product_preview']['value'])) {
                    $images[] = [
                        'featured'  =>  true,
                        'url'       =>  $entry['product_preview']['value'],
                    ];
                }

                $data = [
                    'product_type'                  =>      'product',
                    'name'                          =>      $entry['product_name']['value'],
                    'sku'                           =>      ($entry['product_sku']['value'] ?? $randomValue),
                    'barcode'                       =>      ($entry['product_barcode']['value'] ?? $randomValue),
                    'category_id'                   =>      $category->id,
                    'units'                         =>      [
                        'unit_group'                =>      $unit->group->id,
                        'selling_group'             =>      [
                            [
                                'unit_id'               =>      $unit->id,
                                'preview_url'           =>      ($entry['product_unit_preview']['value'] ?? ''),
                                'sale_price_edit'       =>      ($entry['product_price']['value'] ?? 0),
                                'wholesale_price_edit'  =>      ($entry['product_wholesale_price']['value'] ?? 0),
                            ],
                        ],
                    ],
                    'tax_group_id'                  =>      0,
                    'status'                        =>      $entry['product_status'][ 'value' ] ?? ( $form['product_status'] ?? 'unavailable' ),
                    'type'                          =>      $entry['product_type'][ 'value' ] ?? ( $form['product_type'] ?? 'dematerialized' ),
                    'stock_management'              =>      $entry['product_stock_management']['value'] ?? 'disabled',
                    'tax_type'                      =>      $entry['product_tax_type'][ 'value' ] ?? ( $form['product_tax_type'] ?? 'inclusive' ),
                    'barcode_type'                  =>      $entry['product_barcode_type'][ 'value' ] ?? ( $form['product_barcode_type'] ?? 'code128' ),
                    'images'                        =>      $images,
                ];

                if ($product instanceof Product) {
                    $localResult = $this->productService->update($product, $data);
                    $entry[ 'product' ] = $localResult['data']['product'];

                    $result['success'][] = $localResult;
                } else {
                    $localResult = $this->productService->create($data);
                    $entry[ 'product' ]     =   $localResult['data']['product'];

                    $result['success'][] = $localResult;
                }
            } catch (Exception $exception) {
                throw $exception;
                $result['failures'][] = $exception->getMessage();
            }

            return $entry;
        });

        return [
            'status'    =>  'success',
            'message'   =>  sprintf(
                __m( '%s products were imported, %s failed', 'NsBulkImporter' ),
                count( $result[ 'success' ] ),
                count( $result[ 'failures' ] ),
            ),
            'data'  =>  compact( 'entries', 'result' ),
        ];
    }

    /**
     * Create customers using provided entries
     *
     * @param  array  $entries
     * @param  array  $form
     * @return array response
     */
    public function createCustomer($entries, $form)
    {
        $result = [];

        foreach ($entries as $entry) {
            $address = [
                'name'          =>  $entry['name']['value'],
                'surname'       =>  $entry['surname']['value'] ?? '',
                'email'         =>  $entry['email']['value'] ?? '',
                'address_1'     =>  $entry['address_1']['value'] ?? '',
                'address_2'     =>  $entry['address_2']['value'] ?? '',
                'country'       =>  $entry['country']['value'] ?? '',
                'city'          =>  $entry['city']['value'] ?? '',
                'pobox'         =>  $entry['pobox']['value'] ?? '',
                'company'       =>  $entry['company']['value'] ?? '',
            ];

            $form = [
                'name'      =>  $entry['name']['value'],
                'email'     =>  $entry['email']['value'] ?? '',
                'surname'   =>  $entry['surname']['value'] ?? '',
                'gender'    =>  $entry['gender']['value'] ?? '',
                'pobox'     =>  $entry['poxbox']['value'] ?? '',
                'group_id'  =>  $entry['group_name']['id'] ?? '',
                'address'   =>  [
                    'shipping'  =>  $address,
                    'billing'   =>  $address,
                ],
            ];

            $customer = Customer::where('email', $entry['email']['value'])->first();

            /**
             * if a similar customer exist with the defined
             * email, that customer will be updated.
             */
            if ($customer instanceof Customer) {
                $result[] = $this->customerService->update($customer->id, $form);
            } else {
                $result[] = $this->customerService->create($form);
            }
        }

        return [
            'status'    =>  'success',
            'message'   =>  __('The customers has been imported.'),
            'data'      =>  compact('result'),
        ];
    }

    public function importOrders( $entries, $form )
    {
        /**
         * @var OrdersService
         */
        $ordersService  =   app()->make( OrdersService::class );

        $orders     =   collect( $entries )->groupBy( 'order_code.value' );

        $results    =   $orders->map( function( $products, $orderCode ) use ( $ordersService, $form ) {
            $newProducts   =   collect( $products )->map( function( $orderProduct ) use ( $form ) {
                $quantity   =   $orderProduct[ 'product_quantity' ][ 'value' ];
                $unitPrice  =   ns()->currency->define( $orderProduct[ 'product_price' ][ 'value' ] )->getRaw();

                return [
                    'name'                  =>  $orderProduct[ 'product' ][ 'name' ],
                    'product_id'            =>  $orderProduct[ 'product' ][ 'id' ],
                    'product_category_id'   =>  $orderProduct[ 'product' ][ 'category_id' ],
                    'total_price'           =>  ns()->currency->define( $unitPrice )->multipliedBy( $quantity )->getRaw(),
                    'unit_id'               =>  $form[ 'unit_id' ],
                    'quantity'              =>  $quantity,
                    'discount'              =>  0, // but can be set on column
                    'discount_type'         =>  isset( $orderProduct[ 'product_discount' ][ 'value' ] ) ? 'percentage' : 'flat', // can be set on the import
                    'discount_percentage'   =>  ns()->currency->define( $orderProduct[ 'product_discount' ][ 'value' ] ?? 0 )->getRaw(),
                    'unit_price'            =>  $unitPrice,
                    'unit_quantity_id'      =>  collect( $orderProduct[ 'product' ][ 'unit_quantities' ] )->filter( fn( $unitQuantity ) => $unitQuantity[ 'unit_id' ] === $form[ 'unit_id' ] )->first()[ 'id' ]
                ];
            });

            $payment                =   PaymentType::first();
            $shipping               =   $products[0][ 'order_shipping' ][ 'value' ] ?? 0;
            $discount               =   $products[0][ 'order_discount' ][ 'value' ] ?? 0;
            $discountPercentage     =   $products[0][ 'order_discount_rate' ][ 'value' ] ?? 0;
            $isPercentage           =   isset( $products[0][ 'order_discount_rate' ][ 'value' ] );
            $subTotal               =   $newProducts->map( fn( $product ) => $product[ 'total_price' ] )->sum();
            $discountValueFromRate  =   ns()->currency->define( $subTotal )->multipliedBy( $discountPercentage )->dividedBy(100)->getraw();
            $total                  =   ns()->currency
                ->define( $subTotal )
                ->additionateBy( $shipping )
                ->subtractBy( $discount )
                ->subtractBy( $isPercentage ? $discountValueFromRate : 0 )
                ->getRaw();

            return $ordersService->create([
                'code'  =>  $orderCode,
                'customer_id'   =>  $products[0][ 'customer' ][ 'id' ],
                'products'  =>  $newProducts->toArray(),
                'discount'  =>  $discount,
                'discount_percentage' => $discountPercentage,
                'discount_type' => $isPercentage ? 'percentage' : 'flat',
                'type'  => [
                    'identifier'    =>  $products[0][ 'order_type' ][ 'value' ] ?? 'takeaway',
                ],
                'created_at'    =>  ! empty( $products[0][ 'order_date' ] ) ? Carbon::parse( $products[0][ 'order_date' ][ 'value' ] )->toDateTimeString() : null,
                'shipping' => ns()->currency->define( $products[0][ 'order_shipping' ][ 'value' ] )->getRaw() ?? 0,
                'author'    =>  $form[ 'cashier_id' ] ?? Auth::id(),
                'tax_group_id'  =>  $form[ 'tax_group_id' ],
                'tax_type'  =>  $form[ 'tax_type' ],
                'payments'  =>  [
                    [
                        'identifier'    =>  $form[ 'payment_identifier' ] ?? $payment->identifier,
                        'label' => PaymentType::where( 'identifier', $form[ 'payment_identifier' ] ?? null )->first()?->label ?: $payment->label,
                        'value' =>  $total
                    ]
                ]
            ]);
        });

        return [
            'status'    =>  'success',
            'message'   =>  sprintf(
                __m( '%s order(s) were imported, %s order(s) failed', 'NsBulkImporter' ),
                collect( $results )->filter( fn( $order ) => $order[ 'status' ] === 'success' )->count(),
                collect( $results )->filter( fn( $order ) => $order[ 'status' ] === 'failed' )->count()
            )
        ];
    }

    /**
     * Create providers
     *
     * @param  array  $entries
     */
    public function createProviders($entries)
    {
        collect($entries)->map(function ($entry) {
            return [
                'name'          =>  $entry['name']['value'] ?? __('Undefined Provider'),
                'surname'       =>  $entry['surname']['value'] ?? null,
                'phone'         =>  $entry['phone']['value'] ?? null,
                'email'         =>  $entry['email']['value'] ?? null,
                'address_1'     =>  $entry['address_1']['value'] ?? null,
                'address_2'     =>  $entry['address_2']['value'] ?? null,
                'description'   =>  $entry['description']['value'] ?? null,
            ];
        })->each(function ($provider) {
            $this->providerService->create($provider);
        });
    }

    public function importOptionFile($file)
    {
        if (! is_file($file)) {
            throw new Exception(sprintf(__('Unable to locate the file %s'), $file));
        }

        $options = json_decode(file_get_contents($file));

        foreach ($options as $key => $value) {
            $this->optionsService->set($key, $value);
        }

        return [
            'status'    =>  'success',
            'message'   =>  __('The options has been successfully imported.'),
        ];
    }

    public function procureProductsFromOrders( $entries, $form )
    {
        /**
         * @var BarcodeService
         */
        $barcodeService     =   app()->make( BarcodeService::class );

        $products   =   collect( $entries )->groupBy( 'product.id' );

        $productsToProcure  =   $products->mapWithKeys( function( $product ) use ( $barcodeService ) {
            $product    =   $product->first();

            return [
                $product[ 'product' ][ 'sku' ]  =>  $product
            ];
        });

        /**
         * we'll now try to summarize all the quantities
         */
        $productsToProcure  =   $productsToProcure->map( function( $productToProcure, $key ) use ( $products ) {
            $totalQuantities    =   $products->filter( function( $product ) use ( $key ) {

                return $product->first()[ 'product' ][ 'sku' ] === ( string ) $key;

            })->map( fn( $product ) => $product->first()[ 'product_quantity' ][ 'value' ] )->sum();

            $productToProcure[ 'quantity' ]     =   $totalQuantities;

            return $productToProcure;
        });

        /**
         * @var ProcurementService
         */
        $procurementService     =   app()->make( ProcurementService::class );

        $procurement     =   $procurementService->create([
            'name'  =>  sprintf( __m( 'Imported Procurement %s', 'NsBulkImporter' ), ns()->date->getNowFormatted() ),
            'general'   =>  [
                'provider_id'   =>  Provider::first()->id,
                'invoice_reference' =>  Str::uuid(),
                'delivery_status'   =>  Procurement::DELIVERED,
            ],
            'products'  =>  $productsToProcure->map( function( $product ) use ( $form ) {
                $purchasePrice  =   isset( $product[ 'product_purchase_price' ] ) ? $product[ 'product_purchase_price' ][ 'value' ] : 0;
                return [
                    'name'  =>  $product[ 'product' ][ 'name' ],
                    'product_id'    =>  $product[ 'product' ][ 'id' ],
                    'unit_id'   =>  $form[ 'unit_id' ],
                    'purchase_price'    =>  $purchasePrice,
                    'quantity'  =>  $product[ 'quantity' ],
                    'tax_group_id'  =>  0,
                    'tax_type'  =>  'inclusive', // by default
                    'tax_value' =>  0,
                    'gross_purchase_price' =>  0,
                    'net_purchase_price' =>  0,
                    'tax_value' =>  0,
                    'total_purchase_price'  =>  ns()->currency->define( $purchasePrice )->multiplyBy( $product[ 'quantity' ])->getRaw(),
                ];
            })
        ]);

        return [
            'status'    =>  'success',
            'message'   =>  __m( 'The procurement was successful', 'NsBulkImporter' ),
            'data'      =>  compact( 'procurement' )
        ];
    }
}
