<?php

namespace Modules\NsMultiStore\Services;

use App\Classes\Hook;
use App\Console\Commands\ModuleMigrations;
use App\Models\Migration;
use App\Models\Role;
use App\Models\User;
use App\Services\ModulesService;
use App\Services\Options;
use App\Services\Setup;
use Exception;
use Illuminate\Database\Migrations\Migration as MigrationsMigration;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Storage;
use Illuminate\Support\Str;
use Jackiedo\DotenvEditor\Facades\DotenvEditor;
use Modules\NsMultiStore\Events\MultiStoreTablesCreatedEvent;
use Modules\NsMultiStore\Events\MultiStoreAfterDeletedEvent;
use Modules\NsMultiStore\Exceptions\NotAllowedStoreAccessException;
use Modules\NsMultiStore\Models\Store;
use Modules\NsMultiStore\Models\StoreMigration;

class StoresService
{
    private $isMultiStore = false;

    private $currentStore;

    public $ignored_tables = [
        'nexopos_users',
        'nexopos_roles',
        'nexopos_permissions',
        'nexopos_role_permission',
        'nexopos_stores',
    ];

    public function isMultiStore()
    {
        return $this->currentStore instanceof Store || $this->isMultiStore;
    }

    public function setStore(Store $store)
    {
        $this->currentStore = $store;
        
        /**
         * let's keep a reference to the root
         * settings accessible on each stores.
         */
        ns()->rootOption = ns()->option;

        /**
         * we need to reset the option used on ns() helper to ensure
         * it use the new initialiazed option object.
         */
        app()->singleton( Options::class, function() {
            return ns()->option = new Options;
        });
    }

    public function prefixTable( $table )
    {
        $table = trim($table);

        if ($this->isMultiStore() && ! in_array(trim($table), $this->ignored_tables)) {
            return 'store_'. $this->getCurrentStore()->id.'_'.$table;
        }

        return $table;
    }

    /**
     * @deprecated?
     */
    public function prefixTableName($table)
    {
        if ($this->isMultiStore()) {
            return 'store_'.$table;
        }

        return $table;
    }

    public function getCurrentStore()
    {
        return $this->currentStore;
    }

    public function current()
    {
        return $this->currentStore;
    }

    /**
     * get available migration for
     * a defined store
     * @param Store $store
     * @return Collection
     */
    public function getMigrations( Store $store ): Collection
    {
        /**
         * let's first purge the migration
         * that shouldn't be migrated
         */
        ns()->purgeMissingMigrations();

        /**
         * We'll now pull all the migration
         * file and execute them.
         */
        return collect(
            [
                ...$this->getModuleMigrations(),
                ...$this->getSystemMigrations(),
            ]
        )->diff(
            $this->getExecutedMigration($store)
                ->map(fn (StoreMigration $migration) => $migration->file)
        );
    }

    private function triggerMethod(Store $store, $className, $method)
    {
        /**
         * provide the default prefix
         * that will apply to all table created
         */
        Hook::addFilter('ns-table-prefix', fn ($prefix) => 'store_'.$store->id.'_'.$prefix);

        /**
         * For laravel 9 and anonymous migration
         * we need to check the instance, to resolve it or not.
         */
        if ($className instanceof MigrationsMigration) {
            $classObject = $className;
        } else {
            $classObject = new $className;
        }

        /**
         * we'll now check if that class explicitely allow
         * execution on multistore or not.
         */
        if (method_exists($classObject, 'runOnMultiStore') && $classObject->runOnMultiStore() === false) {
            return;
        }

        $classObject->$method();

        Hook::removeAllFilters('ns-table-prefix');
    }

    public function executeSingleMigration(Store $store, $response, $file)
    {
        if ($response['status'] === 'success') {
            $className = $response['data']['className'];
            $this->triggerMethod($store, $className, 'up');
        }
    }

    public function createStoreTables(Store $store)
    {
        Hook::addFilter( 'ns-table-name', fn( $table ) => 'store_' . $store->id . '_' . $table );
        Hook::addFilter( 'ns-model-table', fn( $table ) => 'store_' . $store->id . '_' . $table );
        
        /**
         * This will execute first system
         * migrations
         */
        collect([
            ...$this->getSystemMigrations(),
        ])->each(fn ($file) => $this->triggerFile($store, $file, 'up'));

        /**
         * This will then execute
         * module migrations
         */
        collect([
            ...$this->getModuleMigrations(),
        ])->each(fn ($file) => $this->triggerFile($store, $file, 'up'));

        /**
         * We'll create the default
         * payments type
         */
        ( new Setup )->createDefaultPayment( User::find( $store->author ) );

        $options    =   new Options;
        $options->set( 'ns_pos_order_types',  ['takeaway', 'delivery']);

        $this->refreshStatefulDomains();

        /**
         * we want to remove the previously
         * defined table name.
         */
        Hook::removeAllFilters( 'ns-table-name' );
        Hook::removeAllFilters( 'ns-model-table' );

        /**
         * when a store table are being created
         * we'll deploy new event
         */
        MultiStoreTablesCreatedEvent::dispatch($store);
    }

    public function refreshStatefulDomains()
    {
        $stores = Store::status(Store::STATUS_OPENED)->get();
        $currentDomains = explode(',', env('SANCTUM_STATEFUL_DOMAINS', ''));

        $domains = $stores->map(function ($store) {
            return $this->getStoreHost($store);
        })->filter(fn ($domain) => ! in_array($domain, $currentDomains))->toArray();

        $finalDomains = array_merge($currentDomains, $domains ?: [], [$this->baseDomainName()]);

        $env = DotenvEditor::load();
        $env->setKey('SANCTUM_STATEFUL_DOMAINS', implode(',', $finalDomains));
        $env->save();

        return [
            'status'    =>  'success',
            'message'   =>  __m('The stateful domains has been refreshed.', 'NsMultiStore'),
        ];
    }

    /**
     * Will detect the migration class
     * and return wether it's from a module or not
     *
     * @param  string  $file
     * @return array $details
     */
    public function getMigrationFileDetails($file)
    {
        $info = pathinfo($file);
        $filename = $info['filename'];
        $className = 'Modules\\'.str_replace('/', '\\', $info['dirname']).'\\'.Str::studly($filename);

        if (class_exists($className)) {
            return [
                'className' =>  $className,
                'module'    =>  explode('/', $info['dirname'])[0],
                'path'      =>  base_path('modules/'.$file),
            ];
        } else {

            /**
             * fetch all system migration
             * and return full path file and name
             */
            $allSystemMigration = collect(Storage::disk('ns')->allFiles('database/migrations'))->mapWithKeys(function ($file) {
                $info = pathinfo($file);

                return [$info['filename'] => $file];
            });

            /**
             * let's check if the file that is provided
             * actually belongs to the system.
             */
            if (in_array($file, $allSystemMigration->keys()->toArray())) {
                $filePath = base_path($allSystemMigration[$file]);

                if (is_file($filePath)) {
                    /**
                     * extract classname from file
                     */
                    $className = Str::studly(collect(explode('_', $file))
                        ->splice(4)
                        ->join('_'));

                    return [
                        'className' =>  $className,
                        'module'    =>  null,
                        'path'      =>  $filePath,
                    ];
                } else {
                    throw new Exception(sprintf(__m('The file "%s" doesn\'t exists', 'NsMultiStore'), $file));
                }
            }

            /**
             * The file wasn't found. Probably deleted on the filesystem.
             * We don't need to run that and then we'll remove that from the executed migrations.
             */
            Migration::where( 'migration', $file )->delete();

            return [
                'className' =>  null,
                'module'    =>  null,
                'path'      =>  null,
            ];
        }   
    }

    /**
     * Trigger a specific migration file
     * by detecting wether it's from a module or a core migration
     *
     * @param  Store  $store
     * @param  string  $file
     * @param  string  $method
     * @return array $response
     */
    public function triggerFile(Store $store, $file, $method): array
    {
        /**
         * @param  string  $module
         * @param  string  $className
         * @param  string  $path
         */
        extract($this->getMigrationFileDetails($file));

        /**
         * in case nothing valid is returned, we might be having a declared
         * migration which file no longer exists. 
         * Therefore the file can't be triggered.
         */
        if( empty( $className ) && empty( $path ) ) {
            return [
                'status'    =>  'info',
                'message'   =>  __m( 'The migration file is missing.', 'NsMultiStore' ),
            ];
        }

        /**
         * let's include the file before 
         * triggering that
         */
        if( ! class_exists( $className ) && ! empty( $path ) ) {
            $object      =   include( $path );

            if ($object instanceof MigrationsMigration) {
                $className = $object;
            } else {
                $className = new $className;
            }
        }

        $this->triggerMethod($store, $className, $method);

        $storeMigration = new StoreMigration;
        $storeMigration->module = $module;
        $storeMigration->store_id = $store->id;
        $storeMigration->file = $file;
        $storeMigration->save();

        return [
            'status'    =>  'success',
            'message'   =>  __m('The migration has been successfully executed.', 'NsMultiStore'),
            'data'      =>  [
                'migration'     =>  $storeMigration,
            ],
        ];
    }

    public function uninstallStore($store)
    {
        Hook::addFilter( 'ns-table-name', fn( $table ) => 'store_' . $store->id . '_' . $table );

        $files = collect(Storage::disk('ns')->files('/database/migrations/create-tables'))
            ->mapWithKeys(function ($file) {
                $fileName = pathinfo($file)['filename'];

                return [
                    $fileName   =>  $file,
                ];
            });

        /**
         * Get store migration files.
         */
        $migrations = StoreMigration::where('store_id', $store->id)->get();

        $migrations->each(function ($migration) use ($files, $store) {
            if ($migration->module === null) {
                if (isset($files[$migration->file]) && is_file(base_path($files[$migration->file]))) {

                    /**
                     * detecting the class name from the file.
                     * This use the migration file name set to file to
                     * extract the class name.
                     */
                    $className      =   collect( explode( '_', $migration->file ) )->skipUntil( function( $value, $index ) {
                        return $index === 4;
                    })->map( function( $part ) {
                        $newPart    =   explode( '.', $part );
                        return ucwords( $newPart[0] );
                    })->join('');

                    if ( ! class_exists( $className ) ) {
                        /**
                         * Include core file for the migration.
                         */
                        $object     =   include( base_path( $files[ $migration->file ] ) );

                        if ( $object instanceof MigrationsMigration ) {
                            $className  =   $object;
                        } 
                    } 
                    
                    $this->triggerMethod( $store, $className, 'down' );
                }

            } else {

                /**
                 * detect the class name from the file.
                 * PSR4 Standard must apply the files
                 */
                $className = Str::studly(pathinfo($migration->file)['filename']);

                if (is_file(base_path('modules'.DIRECTORY_SEPARATOR.$migration->file))) {
                    /**
                     * include the file from the location
                     * stored on the migration.
                     */
                    include_once base_path('modules'.DIRECTORY_SEPARATOR.$migration->file);

                    /**
                     * trigger the down method
                     * to reverse all migration
                     */
                    $this->triggerFile($store, $migration->file, 'down');
                }
            }
        });

        if (is_dir(storage_path('app/public/store_'.$store->id))) {
            Storage::disk('public')->deleteDirectory('store_'.$store->id);
        }

        /**
         * Having the store deleted, we no more
         * need the migration created for that store
         */
        StoreMigration::where('store_id', $store->id)->delete();

        Hook::removeAllFilters( 'ns-table-name' );
    }

    public function dismantleStore(Store $store)
    {
        $this->uninstallStore($store);

        $store->delete();

        MultiStoreAfterDeletedEvent::dispatch( $store );
    }

    public function unsetStore()
    {
        $this->currentStore = null;

        /**
         * we need to bind one more time the Options to point
         * to the root options
         */
        app()->singleton( Options::class, fn() => null );
    }

    public function getStoreURL(Store $store)
    {
        return $this->baseProtocol().$this->getStoreHost($store);
    }

    public function getStoreHost(Store $store)
    {
        return $store->slug.'.'.$this->baseDomainName();
    }

    public function getOpened()
    {
        return Store::where('status', Store::STATUS_OPENED)->get();
    }

    /**
     * Will get opened store
     * to which the logged user has access to
     *
     * @return Collection<Store> a collection of stores.
     */
    public function getOpenedAccessibleStores()
    {
        $stores = Store::where('status', Store::STATUS_OPENED)->get();
        $roles = Auth::user()->roles;

        return $stores->filter(function ($store) use ($roles) {
            $decoded = json_decode($store->roles_id);

            return $roles->filter(fn ($role) => in_array($role->id, $decoded))->count() > 0;
        })->values();
    }

    /**
     * Will update all the store available for
     * a specific role
     *
     * @param  array|int  $roleId
     * @return array
     */
    public function countRoleStoresUsingID($roleId)
    {
        if (is_array($roleId)) {
            $result = [];

            foreach ($roleId as $_roleId) {
                $result[] = $this->countRoleStoresUsingID($_roleId);
            }

            return [
                'status'    =>  'info',
                'message'   =>  __m('The store count has been updated.', 'NsMultiStore'),
                'data'      =>  compact('result'),
            ];
        } else {
            $role = ! $roleId instanceof Role ? Role::find($roleId) : $roleId;
            $stores = Store::where('status', Store::STATUS_OPENED)->get();

            $role->total_stores = $stores->filter(function ($store) use ($role) {
                $decoded = (array) json_decode($store->roles_id, true);

                return in_array($role->id, $decoded);
            })->count();

            $role->save();

            return [
                'status'    =>  'success',
                'message'   =>  __m('The role has been updated.', 'NsMultiStore'),
                'data'      =>  compact('role'),
            ];
        }
    }

    public function getModuleMigrations()
    {
        $module = app()->make(ModulesService::class);

        /**
         * the current module migration
         * shouldn't apply to the sub stores.
         */
        $migrationFiles = collect($module->getEnabled())->filter(function ($module) {
            return $module['namespace'] !== 'NsMultiStore';
        })->map(function ($module) {
            return $module['all-migrations'];
        })->flatten();

        return $migrationFiles;
    }

    /**
     * Return executed migration for a specifc
     * store
     *
     * @param  Store  $store
     * @return Collection
     */
    public function getExecutedMigration(Store $store)
    {
        return StoreMigration::forStore($store)
            ->get();
    }

    /**
     * Return a list of system migration
     *
     * @return Collection $migrations
     */
    public function getSystemMigrations()
    {
        /**
         * here are table that should be
         * created on sub stores.
         */
        return Migration::where( 'type', 'create' )->orderBy('migration', 'asc')
            ->get()
            ->filter(function ($file) {
                extract($this->getMigrationFileDetails($file->migration));

                if ( empty( $className ) && empty( $path ) ) {
                    return false;
                }

                /**
                 * @param  string  $className
                 * @param  string  $module
                 * @param  string  $path
                 */
                if ( ! class_exists( $className ) ) {
                    $object = include $path;
                } else {
                    $object = null;
                }

                if ( ! $object instanceof MigrationsMigration ) {
                    $object = new $className;
                }

                return ! method_exists($object, 'runOnMultistore') || $object->runOnMultiStore();
            })
            ->map(fn ($migration) => $migration->migration);
    }

    /**
     * Defines the store routes
     *
     * @return void
     */
    public function defineStoreRoutes($callback)
    {
        $this->isMultiStore = true;
        $callback();
        $this->isMultiStore = false;
    }

    /**
     * Returns wether the sub domains are
     * currently enabled on the store
     *
     * @return bool
     */
    public function subDomainsEnabled()
    {
        return ns()->store->isMultiStore() ?
            ns()->rootOption->get('nsmultistore-subdomain', 'disabled') === 'enabled' :
            ns()->option->get('nsmultistore-subdomain', 'disabled') === 'enabled';
    }

    /**
     * Returns the actual base domain
     *
     * @param  string  $domain
     */
    public function baseDomainName()
    {
        $domain = pathinfo(env('APP_URL'));

        return $domain['filename'].(isset($domain['extension']) ? '.'.$domain['extension'] : '');
    }

    /**
     * Returns the current protocol used
     * on the store
     *
     * @return string $store
     */
    public function baseProtocol()
    {
        $domain = pathinfo(env('APP_URL'));

        return $domain['dirname'].'//';
    }

    /**
     * Will check if the currently logged user
     * has the right to access the store
     *
     * @return void
     */
    public function checkStoreAccessibility()
    {
        if (
            ns()->store->subDomainsEnabled() &&
            ns()->store->isMultiStore()
        ) {
            $storeRoles = (array) json_decode(Store::current()->roles_id, true);
            $roles = Auth::user()->roles;

            $hasAccess = $roles->filter(function ($role) use ($storeRoles) {
                return in_array($role->id, $storeRoles);
            });

            if ($hasAccess->isEmpty()) {
                throw new NotAllowedStoreAccessException(__m('You\'re not allowed to access to this store. If the problem persist, please contact the administrator.', 'NsMultiStore'));
            }
        }
    }
}
