<?php
namespace Modules\NsStockTransfers\Crud;

use App\Crud\ProductUnitQuantitiesCrud;
use Illuminate\Support\Facades\Auth;
use Illuminate\Http\Request;
use App\Services\CrudService;
use App\Services\Users;
use App\Exceptions\NotAllowedException;
use App\Models\ProductUnitQuantity;
use App\Models\User;
use App\Services\CrudEntry;
use App\Services\Helper;
use App\Services\ProductService;
use TorMorten\Eventy\Facades\Events as Hook;
use Exception;
use Modules\NsMultiStore\Models\Store;
use Modules\NsStockTransfers\Events\TransactionAfterProcessedEvent;
use Modules\NsStockTransfers\Models\StockTransfer;
use Modules\NsStockTransfers\Models\StockTransferProduct;

class StockTransferCrud extends CrudService
{
    /**
     * define the base table
     * @param  string
     */
    protected $table      =   'nexopos_stock_transfers';

    /**
     * default slug
     * @param  string
     */
    protected $slug   =   'stock-transfers';

    /**
     * Define namespace
     * @param  string
     */
    protected $namespace  =   'ns.stocktransfers-transfers';

    /**
     * Model Used
     * @param  string
     */
    protected $model      =   StockTransfer::class;

    /**
     * Define permissions
     * @param  array
     */
    protected $permissions  =   [
        'create'    =>  'stocktransfers.create.transaction',
        'read'      =>  'stocktransfers.read.transaction',
        'update'    =>  'stocktransfers.update.transaction',
        'delete'    =>  'stocktransfers.delete.transaction',
    ];

    /**
     * Adding relation
     * @param  array
     */
    public $relations   =  [
        [ 'nexopos_users as user', 'user.id', '=', 'nexopos_stock_transfers.author' ],
        'leftJoin'  =>  [
            [ 'nexopos_stores as from_store', 'from_store.id', '=', 'nexopos_stock_transfers.from_store' ],
            [ 'nexopos_stores as to_store', 'to_store.id', '=', 'nexopos_stock_transfers.to_store' ],
        ]
    ];

    /**
     * all tabs mentionned on the tabs relations
     * are ignored on the parent model.
     */
    protected $tabsRelations    =   [
        // 'tab_name'      =>      [ YourRelatedModel::class, 'localkey_on_relatedmodel', 'foreignkey_on_crud_model' ],
    ];

    /**
     * Pick
     * Restrict columns you retreive from relation.
     * Should be an array of associative keys, where 
     * keys are either the related table or alias name.
     * Example : [
     *      'user'  =>  [ 'username' ], // here the relation on the table nexopos_users is using "user" as an alias
     * ]
     */
    public $pick        =   [
        'user'     =>  [ 'username' ]
    ];

    /**
     * Define where statement
     * @var  array
    **/
    protected $listWhere    =   [];

    /**
     * Define where in statement
     * @var  array
     */
    protected $whereIn      =   [];

    /**
     * Fields which will be filled during post/put
     */
        public $fillable    =   [];

    /**
     * @var ProductService
     */
    protected $productService;

    /**
     * Define Constructor
     * @param  
     */
    public function __construct()
    {
        parent::__construct();

        $this->productService       =   app()->make( ProductService::class );

        Hook::addFilter( $this->namespace . '-crud-actions', [ $this, 'setActions' ], 10, 2 );
    }

    /**
     * Return the label used for the crud 
     * instance
     * @return  array
    **/
    public function getLabels()
    {
        return [
            'list_title'            =>  __( 'Stock Transfers List' ),
            'list_description'      =>  __( 'Display all stock transfers.' ),
            'no_entry'              =>  __( 'No stock transfers has been registered' ),
            'create_new'            =>  __( 'Add a new stock transfer' ),
            'create_title'          =>  __( 'Create a new stock transfer' ),
            'create_description'    =>  __( 'Register a new stock transfer and save it.' ),
            'edit_title'            =>  __( 'Edit stock transfer' ),
            'edit_description'      =>  __( 'Modify  Stock Transfer.' ),
            'back_to_list'          =>  __( 'Return to Stock Transfers' ),
        ];
    }

    /**
     * Check whether a feature is enabled
     * @return  boolean
    **/
    public function isEnabled( $feature ): bool
    {
        return false; // by default
    }

    /**
     * Fields
     * @param  object/null
     * @return  array of field
     */
    public function getForm( $entry = null ) 
    {
        $stores      =   Store::status( Store::STATUS_OPENED )
            ->where( 'id', '<>', ns()->store->current()->id )
            ->get()
            ->map( function( $store ) {
                return [
                    'label' =>  $store->name,
                    'value' =>  $store->id
                ];
            });

        return [
            'main' =>  [
                'label'         =>  __( 'Name' ),
                'name'          =>  'name',
                'value'         =>  $entry->name ?? '',
                'description'   =>  __( 'Provide a name to the resource.' )
            ],
            'products'          =>  $entry !== null ? $entry->products->map( function( StockTransferProduct $stockTransferProduct ) {
                $stockTransferProduct->unit_quantity     =   $this->productService->getUnitQuantity( 
                    $stockTransferProduct->product_id, $stockTransferProduct->unit_id
                );

                $stockTransferProduct->unit_quantities   =   $this->productService->getProductUnitQuantities( $stockTransferProduct->product );

                return $stockTransferProduct;
            }) : [],
            'tabs'  =>  [
                'general'   =>  [
                    'label'     =>  __( 'General' ),
                    'fields'    =>  [
                        [
                            'type'          =>  'select',
                            'name'          =>  'to_store',
                            'label'         =>  __( 'To Store' ),
                            'description'   =>  __( 'Select the destination store where the stock transfer goes.' ),
                            'validation'    =>  'required',
                            'value'         =>  $entry->to_store ?? '',
                            'options'       =>  $stores,
                        ], [
                            'type'          =>  'select',
                            'name'          =>  'status',
                            'label'         =>  __( 'Request Status' ),
                            'description'   =>  __( 'New requests starts with a pending status.' ),
                            'validation'    =>  'required',
                            'value'         =>   $entry->status ?? 'draft',
                            'options'       =>  Helper::kvToJsOptions([
                                StockTransfer::STATUS_DRAFT     =>  __m( 'Draft', 'NsStockTransfers' ),
                                StockTransfer::STATUS_PENDING   =>  __m( 'Pending', 'NsStockTransfers' ),
                            ]),
                        ], [
                            'type'          =>  'hidden',
                            'name'          =>  'type',
                            'label'         =>  __( 'Request Type' ),
                            'description'   =>  __( 'Select wether the store is requesting or transfering stock.' ),
                            'validation'    =>  'required',
                            'value'         =>   'transfer',
                        ], [
                            'type'          =>  'hidden',
                            'name'          =>  'from_store',
                            'label'         =>  __( 'From Store' ),
                            'description'   =>  __( 'Select the source store where the stock transfer goes.' ),
                            'validation'    =>  'required',
                            'value'         =>  ns()->store->current()->id
                        ], 
                    ]
                ],
                'table'     =>  [
                    'label'     =>  __( 'Products' ),
                    'fields'    =>  [],
                ],
            ]
        ];
    }

    public function hook( $query ): void
    {
        $currentStore   =   ns()->store->current()->id;
        
        $query->orderBy( 'updated_at', 'desc' );

        $query->where( function( $query ) use ( $currentStore ) {
            $query->orWhere( function( $query ) use ( $currentStore ) {
                $query
                    ->where( 'from_store', $currentStore )
                    ->where( 'type', StockTransfer::TYPE_TRANSFER )
                    ->whereIn( 'nexopos_stock_transfers.status', [
                        StockTransfer::STATUS_DRAFT,
                        StockTransfer::STATUS_PENDING,
                        StockTransfer::STATUS_ONGOING,
                        StockTransfer::STATUS_DELIVERED,
                        StockTransfer::STATUS_STOCKED,
                        StockTransfer::STATUS_REJECTED,
                        StockTransfer::STATUS_CANCELED,
                        StockTransfer::STATUS_REVIEW,
                    ]);
            });

            $query->orWhere( function( $query ) use ( $currentStore ) {
                $query
                    ->where( 'to_store', $currentStore )
                    ->where( 'type', StockTransfer::TYPE_TRANSFER )
                    ->whereIn( 'nexopos_stock_transfers.status', [
                        StockTransfer::STATUS_PENDING,
                        StockTransfer::STATUS_STOCKED,
                        StockTransfer::STATUS_CANCELED,
                        StockTransfer::STATUS_REJECTED,
                        StockTransfer::STATUS_REVIEW,
                    ]);
            });

            $query->orWhere( function( $query ) use ( $currentStore ) {
                $query
                    ->where( 'from_store', $currentStore )
                    ->where( 'type', StockTransfer::TYPE_REQUEST )
                    ->whereIn( 'nexopos_stock_transfers.status', [
                        StockTransfer::STATUS_PENDING,
                        StockTransfer::STATUS_ONGOING,
                        StockTransfer::STATUS_DELIVERED,
                        StockTransfer::STATUS_STOCKED,
                        StockTransfer::STATUS_REJECTED,
                        StockTransfer::STATUS_REVIEW,
                    ]);
            });

            $query->orWhere( function( $query ) use ( $currentStore ) {
                $query
                    ->where( 'to_store', $currentStore )
                    ->where( 'type', StockTransfer::TYPE_REQUEST )
                    ->whereIn( 'nexopos_stock_transfers.status', [
                        StockTransfer::STATUS_DRAFT,
                        StockTransfer::STATUS_PENDING,
                        StockTransfer::STATUS_ONGOING,
                        StockTransfer::STATUS_DELIVERED,
                        StockTransfer::STATUS_STOCKED,
                        StockTransfer::STATUS_REJECTED,
                        StockTransfer::STATUS_CANCELED,
                        StockTransfer::STATUS_REVIEW,
                    ]);
            });
        });
    }

    /**
     * We don't want the table name to be hookable
     * as it will run on the warehouse level
     * @return string
     */
    protected function hookTableName( $tableName ): string
    {
        return $tableName;
    }

    /**
     * Filter POST input fields
     * @param  array of fields
     * @return  array of fields
     */
    public function filterPostInputs( $inputs )
    {
        return $inputs;
    }

    /**
     * Filter PUT input fields
     * @param  array of fields
     * @return  array of fields
     */
    public function filterPutInputs( $inputs, StockTransfer $entry )
    {
        return $inputs;
    }

    /**
     * Before saving a record
     * @param  Request $request
     * @return  void
     */
    public function beforePost( $request )
    {
        if ( $this->permissions[ 'create' ] !== false ) {
            ns()->restrict( $this->permissions[ 'create' ] );
        } else {
            throw new NotAllowedException;
        }

        return $request;
    }

    /**
     * After saving a record
     * @param  Request $request
     * @param  StockTransfer $entry
     * @return  void
     */
    public function afterPost( $request, StockTransfer $entry )
    {
        return $request;
    }

    
    /**
     * get
     * @param  string
     * @return  mixed
     */
    public function get( $param )
    {
        switch( $param ) {
            case 'model' : return $this->model ; break;
        }
    }

    /**
     * Before updating a record
     * @param  Request $request
     * @param  object entry
     * @return  void
     */
    public function beforePut( $request, $entry )
    {
        if ( $this->permissions[ 'update' ] !== false ) {
            ns()->restrict( $this->permissions[ 'update' ] );
        } else {
            throw new NotAllowedException;
        }

        return $request;
    }

    /**
     * After updating a record
     * @param  Request $request
     * @param  object entry
     * @return  void
     */
    public function afterPut( $request, $entry )
    {
        return $request;
    }

    /**
     * Before Delete
     * @return  void
     */
    public function beforeDelete( $namespace, $id, $model ) {
        if ( $namespace == 'ns.stocktransfers-transfers' ) {
            /**
             *  Perform an action before deleting an entry
             *  In case something wrong, this response can be returned
             *
             *  return response([
             *      'status'    =>  'danger',
             *      'message'   =>  __( 'You\re not allowed to do that.' )
             *  ], 403 );
            **/
            if ( $this->permissions[ 'delete' ] !== false ) {
                ns()->restrict( $this->permissions[ 'delete' ] );
            } else {
                throw new NotAllowedException;
            }

            if (  $model->status  === StockTransfer::STATUS_ONGOING ) {
                throw new Exception( __( 'Cannot delete a stock transaction that is ongoing.' ) );
            }

            if (  $model->status  === StockTransfer::STATUS_STOCKED ) {
                throw new Exception( __( 'Cannot delete a stock transaction that is already stocked.' ) );
            }

            $model->products->each( function( StockTransferProduct $product ) {
                $product->delete();
            });
        }
    }

    /**
     * Define Columns
     * @return  array of columns configuration
     */
    public function getColumns() {
        return [
            'name'  =>  [
                'label'  =>  __( 'Name' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'status'  =>  [
                'label'  =>  __( 'Status' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'type'  =>  [
                'label'  =>  __( 'Type' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'from_store_name'  =>  [
                'label'         =>  __( 'From Store' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'to_store_name'  =>  [
                'label'         =>  __( 'To Store' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'status'  =>  [
                'label'  =>  __( 'Status' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'created_at'  =>  [
                'label'  =>  __( 'Created At' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ],
            'user_username'  =>  [
                'label'  =>  __( 'Author' ),
                '$direction'    =>  '',
                '$sort'         =>  false
            ]
        ];
    }

    /**
     * Define actions
     */
    public function setActions( CrudEntry $entry )
    {
        // Don't overwrite
        $entry->{ '$checked' }  =   false;
        $entry->{ '$toggled' }  =   false;
        $entry->{ '$id' }       =   $entry->id; 

        $entry->addAction( 'ns.stocktransfers-invoice', [
            'label'     =>  __( 'Invoice' ),
            'namespage' =>  'ns.stocktransfers-invoice',
            'type'      =>  'GOTO',
            'url'       =>  ns()->url( '/dashboard/stock-transfers/invoice/' . $entry->id ),
        ]);

        /**
         * IF the current store is the destination store (to_store)
         * and the request is a transfer, we should provide an option to
         * approve the transfer. 
         * @todo add permission to restrict this action
         */        
        if ( 
            $entry->type === StockTransfer::TYPE_TRANSFER &&
            $entry->status === StockTransfer::STATUS_PENDING
        ) {
            if ( 
                ns()->store->current()->id === $entry->to_store
            ) {

                if ( ns()->rootOption->get( 'ns-stocktransfers-approval', 'no' ) === 'yes' ) {
                    $entry->addAction( 'ns.stocktransfer.review', [
                        'label'     =>  __( 'Review Transfer' ),
                        'namespage' =>  'ns.stocktransfer.review',
                        'type'      =>  'GOTO',
                        'url'       =>  ns()->route( 'ns.stocktransfer.review', [
                            'transaction'   =>  $entry->id
                        ]),
                    ]);
                } else {
                    $entry->addAction( 'ns.stocktransfers-approve', [
                        'label'     =>  __( 'Approve Transfer' ),
                        'namespage' =>  'ns.stocktransfers-approve',
                        'type'      =>  'GET',
                        'url'       =>  ns()->url( '/api/nexopos/v4/stock-transfers/approve/' . $entry->id ),
                        'confirm'   =>  [
                            'message'   =>  __( 'Would you like to approve this transfer ?' )
                        ]
                    ]);
                }
                
                $entry->addAction( 'ns.stocktransfers-reject', [
                    'label'     =>  __( 'Reject Transfer' ),
                    'namespage' =>  'ns.stocktransfers-reject',
                    'type'      =>  'GET',
                    'url'       =>  ns()->url( '/api/nexopos/v4/stock-transfers/reject/' . $entry->id ),
                    'confirm'   =>  [
                        'message'   =>  __( 'Would you like to reject this transfer ?' )
                    ]
                ]);

            } else {
                $entry->addAction( 'ns.stocktransfers-cancel', [
                    'label'     =>  __( 'Cancel Transfer' ),
                    'namespage' =>  'ns.stocktransfers-cancel',
                    'type'      =>  'GET',
                    'url'       =>  ns()->url( '/api/nexopos/v4/stock-transfers/cancel/' . $entry->id ),
                    'confirm'   =>  [
                        'message'   =>  __( 'Would you like to cancel this transfer ?' )
                    ]
                ]);
            }
        }

        if ( 
            $entry->type === StockTransfer::TYPE_REQUEST &&
            $entry->status === StockTransfer::STATUS_PENDING
        ) {
            if ( 
                ns()->store->current()->id === $entry->from_store
            ) {
                $entry->addAction( 'ns.stocktransfers-approve', [
                    'label'     =>  __( 'Approve Request' ),
                    'namespage' =>  'ns.stocktransfers-approve',
                    'type'      =>  'GET',
                    'url'       =>  ns()->url( '/api/nexopos/v4/stock-transfers/approve/' . $entry->id ),
                    'confirm'   =>  [
                        'message'   =>  __( 'Would you like to approve this request ?' )
                    ]
                ]);
                
                $entry->addAction( 'ns.stocktransfers-reject', [
                    'label'     =>  __( 'Reject Request' ),
                    'namespage' =>  'ns.stocktransfers-reject',
                    'type'      =>  'GET',
                    'url'       =>  ns()->url( '/api/nexopos/v4/stock-transfers/reject/' . $entry->id ),
                    'confirm'   =>  [
                        'message'   =>  __( 'Would you like to reject this request ?' )
                    ]
                ]);

                $entry->addAction( 'ns.stocktransfers-edit', [
                    'label'     =>  __( 'Edit Request' ),
                    'namespage' =>  'ns.stocktransfers-edit',
                    'type'      =>  'GOTO',
                    'url'       =>  ns()->url( '/dashboard/' . 'stock-requests' . '/edit/' . $entry->id ),
                ]);

            } else {
                $entry->addAction( 'ns.stocktransfers-cancel', [
                    'label'     =>  __( 'Cancel Request' ),
                    'namespage' =>  'ns.stocktransfers-cancel',
                    'type'      =>  'GET',
                    'url'       =>  ns()->url( '/api/nexopos/v4/stock-transfers/cancel/' . $entry->id ),
                    'confirm'   =>  [
                        'message'   =>  __( 'Would you like to cancel this request ?' )
                    ]
                ]);
            }
        }

        if ( 
            $entry->type === StockTransfer::TYPE_REQUEST &&
            $entry->status === StockTransfer::STATUS_REVIEW
        ) {
            if ( 
                ns()->store->current()->id === $entry->to_store
            ) {
                $entry->addAction( 'ns.stocktransfer.review', [
                    'label'     =>  __( 'Review Request' ),
                    'namespage' =>  'ns.stocktransfer.review',
                    'type'      =>  'GOTO',
                    'url'       =>  ns()->route( 'ns.stocktransfer.review', [
                        'transaction'   =>  $entry->id
                    ])
                ]);
            }
        }

        if ( 
            in_array( $entry->type, [ 
                StockTransfer::TYPE_TRANSFER,
            ]) &&
            $entry->from_store === ns()->store->current()->id &&
            in_array( $entry->status, [
                StockTransfer::STATUS_DRAFT,
                StockTransfer::STATUS_PENDING,
                StockTransfer::STATUS_CANCELED,
                StockTransfer::STATUS_REJECTED
            ])
        ) {
            $entry->addAction( 'edit', [
                'label'         =>      __( 'Edit' ),
                'namespace'     =>      'edit',
                'type'          =>      'GOTO',
                'url'           =>      ns()->url( '/dashboard/' . 'stock-transfers' . '/edit/' . $entry->id )
            ]);
            
            $entry->addAction( 'delete', [
                'label'     =>  __( 'Delete' ),
                'namespace' =>  'delete',
                'type'      =>  'DELETE',
                'url'       =>  ns()->url( '/api/nexopos/v4/crud/ns.stocktransfers-transfers/' . $entry->id ),
                'confirm'   =>  [
                    'message'  =>  __( 'Would you like to delete this ?' ),
                ]
            ]);
        }

        if ( 
            in_array( $entry->type, [ 
                StockTransfer::TYPE_REQUEST,
            ]) &&
            $entry->to_store === ns()->store->current()->id &&
            in_array( $entry->status, [
                StockTransfer::STATUS_DRAFT,
                StockTransfer::STATUS_PENDING,
                StockTransfer::STATUS_CANCELED,
                StockTransfer::STATUS_REJECTED
            ])
        ) {
            $entry->addAction( 'edit', [
                'label'         =>      __( 'Edit' ),
                'namespace'     =>      'edit',
                'type'          =>      'GOTO',
                'url'           =>      ns()->url( '/dashboard/' . 'stock-transfers' . '/edit/' . $entry->id )
            ]);
            
            $entry->addAction( 'delete', [
                'label'     =>  __( 'Delete' ),
                'namespace' =>  'delete',
                'type'      =>  'DELETE',
                'url'       =>  ns()->url( '/api/nexopos/v4/crud/ns.stocktransfers-transfers/' . $entry->id ),
                'confirm'   =>  [
                    'message'  =>  __( 'Would you like to delete this ?' ),
                ]
            ]);
        }

        switch( $entry->type ) {
            case StockTransfer::TYPE_REQUEST:     
                $entry->type = __( 'Request' ); 
            break;
            case StockTransfer::TYPE_TRANSFER:     
                $entry->type = __( 'Transfer' ); 
            break;
        }

        switch( $entry->status ) {
            case StockTransfer::STATUS_PENDING:     
                $entry->status = __( 'Pending' ); 
                $entry->{ '$cssClass' }     =   'info border-b';
            break;
            case StockTransfer::STATUS_STOCKED:    
                $entry->status = __( 'Stocked' ); 
                $entry->{ '$cssClass' }     =   'success border-b';
            break;
            case StockTransfer::STATUS_DRAFT:       
                $entry->status = __( 'Draft' ); 
                $entry->{ '$cssClass' }     =   'border-b';
            break;
            case StockTransfer::STATUS_DELIVERED:   
                $entry->status = __( 'Delivered' ); 
                $entry->{ '$cssClass' }     =   'success border-b';
            break;
            case StockTransfer::STATUS_ONGOING:     
                $entry->status = __( 'Ongoing' ); 
                $entry->{ '$cssClass' }     =   'info border-b';
            break;
            case StockTransfer::STATUS_REJECTED:     
                $entry->status = __( 'Rejected' ); 
                $entry->{ '$cssClass' }     =   'error border-b';
            break;
            case StockTransfer::STATUS_CANCELED:     
                $entry->status = __( 'Canceled' ); 
                $entry->{ '$cssClass' }     =   'warning border-b';
            break;
            case StockTransfer::STATUS_REVIEW:     
                $entry->status = __( 'Awaiting Review' ); 
                $entry->{ '$cssClass' }     =   'info border-b';
            break;
        }    

        return $entry;
    }

    
    /**
     * Bulk Delete Action
     * @param    object Request with object
     * @return    false/array
     */
    public function bulkAction( Request $request ) 
    {
        /**
         * Deleting licence is only allowed for admin
         * and supervisor.
         */

        if ( $request->input( 'action' ) == 'delete_selected' ) {

            /**
             * Will control if the user has the permissoin to do that.
             */
            if ( $this->permissions[ 'delete' ] !== false ) {
                ns()->restrict( $this->permissions[ 'delete' ] );
            } else {
                throw new NotAllowedException;
            }

            $status     =   [
                'success'   =>  0,
                'failed'    =>  0
            ];

            foreach ( $request->input( 'entries' ) as $id ) {
                $entity     =   $this->model::find( $id );
                if ( $entity instanceof StockTransfer ) {
                    try {
                        $this->beforeDelete( $this->namespace, $id, $entity );
                        $entity->delete();
                        $status[ 'success' ]++;
                    } catch( Exception $exception ) {
                        $status[ 'failed' ]++;
                    }
                } else {
                    $status[ 'failed' ]++;
                }
            }
            return $status;
        }

        return Hook::filter( $this->namespace . '-catch-action', false, $request );
    }

    /**
     * get Links
     * @return  array of links
     */
    public function getLinks(): array
    {
        return  [
            'list'      =>  ns()->url( 'dashboard/' . 'stock-transfers' ),
            'create'    =>  ns()->url( 'dashboard/' . 'stock-transfers/create/transfer' ),
            'edit'      =>  ns()->url( 'dashboard/' . 'stock-transfers/edit/transfer/' ),
            'post'      =>  ns()->url( 'api/nexopos/v4/crud/' . 'ns.stocktransfers-transfers' ),
            'put'       =>  ns()->url( 'api/nexopos/v4/crud/' . 'ns.stocktransfers-transfers/{id}' . '' ),
        ];
    }

    /**
     * Get Bulk actions
     * @return  array of actions
    **/
    public function getBulkActions(): array
    {
        return Hook::filter( $this->namespace . '-bulk', [
            [
                'label'         =>  __( 'Delete Selected Groups' ),
                'identifier'    =>  'delete_selected',
                'url'           =>  ns()->route( 'ns.api.crud-bulk-actions', [
                    'namespace' =>  $this->namespace
                ])
            ]
        ]);
    }

    /**
     * get exports
     * @return  array of export formats
    **/
    public function getExports()
    {
        return [];
    }
}