<?php
namespace Modules\NsStockTransfers\Http\Controllers;

use App\Exceptions\NotAllowedException;
use App\Http\Controllers\DashboardController;
use App\Models\Product;
use Exception;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Validator;
use Modules\NsMultiStore\Models\Store;
use Modules\NsStockTransfers\Crud\StockTransferCrud;
use Modules\NsStockTransfers\Http\Requests\StockTransferRequest;
use Modules\NsStockTransfers\Models\StockTransfer;
use Modules\NsStockTransfers\Models\StockTransferProduct;
use Modules\NsStockTransfers\Services\StockTransferService;
use Modules\NsStockTransfers\Settings\GeneralSettings;

class StockTransfersController extends DashboardController
{
    private $stockTransferService;
    
    public function __construct(
        StockTransferService $stockTransferService
    )
    {
        parent::__construct();
        $this->stockTransferService     =   $stockTransferService;    
    }
    
    public function transferList()
    {
        ns()->restrict([ 'stocktransfers.read.transaction' ]);

        return StockTransferCrud::table();
    }

    public function createTransfer()
    {
        ns()->restrict([ 'stocktransfers.create.transaction' ]);

        return $this->view( 'NsStockTransfers::transfers.transfer', [
            'title'         =>  __( 'Create New Transfer' ),
            'description'   =>  __( 'Create a new stock transfer.' )
        ]);
    }

    public function editTransfer( StockTransfer $transaction )
    {
        ns()->restrict([ 'stocktransfers.update.transaction' ]);

        if ( $transaction->type === StockTransfer::TYPE_REQUEST ) {
            return $this->view( 'NsStockTransfers::transfers.request', [
                'title'         =>  __( 'Update Request' ),
                'description'   =>  __( 'Update an existing request.' ),
                'transaction'   =>  $transaction
            ]);
        }

        return $this->view( 'NsStockTransfers::transfers.transfer', [
            'title'         =>  __( 'Update Transfer' ),
            'description'   =>  __( 'Update an existing transfer.' ),
            'transaction'   =>  $transaction
        ]);
    }

    public function settings()
    {
        return GeneralSettings::renderForm();
    }

    public function editStockRequest( StockTransfer $transaction ) 
    {
        ns()->restrict([ 'stocktransfers.update.transaction' ]);

        if ( 
            ! $transaction->status === StockTransfer::STATUS_PENDING && 
            ! (
                Store::current() instanceof Store && 
                $transaction->to_store !== Store::current()
            )
        ) {
            throw new NotAllowedException( __( 'You\'re not allowed to see this page.' ) );
        }

        return $this->view( 'NsStockTransfers::transfers.request', [
            'title'         =>  __( 'Update Request' ),
            'description'   =>  __( 'Update an existing request.' ),
            'transaction'   =>  $transaction
        ]);
    }

    public function deleteProduct( StockTransferProduct $product )
    {
        ns()->restrict([ 'stocktransfers.delete.transaction' ]);

        $product->delete();

        return [
            'status'    =>  'success',
            'message'   =>  __( 'The product has been removed from the transaction.' ),
        ];
    }

    public function createRequest()
    {
        ns()->restrict([ 'stocktransfers.create.transaction' ]);

        return $this->view( 'NsStockTransfers::transfers.request', [
            'title'         =>  __( 'Create New Request' ),
            'description'   =>  __( 'Create a new stock request.' )
        ]);
    }

    public function submitTransaction( StockTransferRequest $request )
    {
        ns()->restrict([ 'stocktransfers.create.transaction' ]);

        return $this->stockTransferService->updateStockTransaction( 
            array_merge(
                $request->input( 'general' ), [
                    'name'      =>  $request->input( 'name' ),
                    'products'  =>  $request->input( 'products' ),
                ]
            )
        );
    }

    public function updateTransaction( StockTransferRequest $request, StockTransfer $transaction )
    {
        ns()->restrict([ 'stocktransfers.update.transaction' ]);

        return $this->stockTransferService->updateStockTransaction(
            array_merge(
                $request->input( 'general' ), [
                    'name'          =>  $request->input( 'name' ),
                    'products'      =>  $request->input( 'products' ),
                ]
            ), 
            $transaction
        );
    }

    public function searchProduct( Request $request, StockTransfer $transaction = null )
    {
        ns()->restrict([ 'stocktransfers.read.transaction' ]);


        $fields         =   $request->input( 'fields' );
        $search         =   $request->input( 'search' );

        $fromStore      =   ( array ) collect( $fields )->filter( fn( $field ) => $field[ 'name' ] === 'from_store' )->first();
        $toStore        =   ( array ) collect( $fields )->filter( fn( $field ) => $field[ 'name' ] === 'to_store' )->first();
        $type           =   ( array ) collect( $fields )->filter( fn( $field ) => $field[ 'name' ] === 'type' )->first();

        $fromStore      =   $fromStore[ 'value' ] ?? null;
        $toStore        =   $toStore[ 'value' ] ?? null;
        $type           =   $type[ 'value' ] ?? null;

        $validator      =   Validator::make( compact( 'fromStore', 'toStore', 'type' ), [
            'fromStore' =>  'required',
            'toStore'   =>  'required',
            'type'      =>  'required',
        ]);

        if ( $validator->fails() ) {
            return response([
                'status'    =>  'failed',
                'message'   =>  __( 'The form is not valid' ),
                'error'     =>  $validator->errors()
            ], 422 );
        }

        if ( $fromStore == $toStore ) {
            throw new Exception( __( 'Source store and destination store cannot be the same.' ) );
        }

        $sourceStore            =   Store::find( $fromStore );
        $destinationStore       =   Store::find( $toStore );

        if ( $type === 'transfer' && $destinationStore->id === ns()->store->current()->id && ! $transaction instanceof StockTransfer ) {
            throw new Exception( 
                sprintf( 
                    __( 'Transfer must only be created by the source store (%s). You can request stock from (%s).' ),
                    $sourceStore->name,
                    $sourceStore->name,
                )
            );
        }

        if ( $type === 'request' && $sourceStore->id === ns()->store->current()->id && ! $transaction instanceof StockTransfer ) {
            throw new Exception( 
                sprintf( 
                    __( 'Request must only be created by the destination store (%s). You can transfer stock to (%s).' ),
                    $destinationStore->name,
                    $destinationStore->name,
                )
            );
        }
        
        /**
         * We'll determine this store as
         * active to pull all products from it.
         */
        Store::switchTo( $sourceStore );

        $products   =   Product::with( 'unit_quantities.unit' )
            ->withStockEnabled()
            ->where( 'name', 'LIKE', "%{$search}%" )
            ->orWhere( 'sku', $search )
            ->orWhere( 'barcode', $search )
            ->limit(10)
            ->get();

        return $products;
    }

    public function rejetTransaction( StockTransfer $transaction )
    {
        if ( $transaction->to_store !== ns()->store->current()->id ) {
            throw new NotAllowedException( __( 'You\'re not allowed to perform this action.' ) );
        }

        return $this->stockTransferService->rejetTransaction( $transaction );
    }

    public function cancelTransaction( StockTransfer $transaction )
    {
        if ( $transaction->from_store !== ns()->store->current()->id ) {
            throw new NotAllowedException( __( 'You\'re not allowed to perform this action.' ) );
        }

        return $this->stockTransferService->cancelTransaction( $transaction );
    }

    public function approveTransaction( StockTransfer $transaction )
    {
        ns()->restrict([ 'stocktransfers.update.transaction']);

        if ( $transaction->status !== StockTransfer::STATUS_PENDING ) {
            throw new NotAllowedException( __( 'The status of the transaction is not valid for an approval.' ) );
        }

        if ( $transaction->to_store !== ns()->store->current()->id && $transaction->type === StockTransfer::TYPE_TRANSFER ) {
            throw new NotAllowedException( __( 'You\'re not allowed to perform this action.' ) );
        }

        if ( $transaction->from_store !== ns()->store->current()->id && $transaction->type === StockTransfer::TYPE_REQUEST ) {
            throw new NotAllowedException( __( 'You\'re not allowed to perform this action.' ) );
        }

        return $this->stockTransferService->approveTransaction( $transaction );
    }

    public function getInvoice( StockTransfer $transaction )
    {
        $transaction->load( 'sourceStore' );
        $transaction->load( 'destinationStore' );
        $transaction->load( 'user' );

        switch( $transaction->status ) {
            case StockTransfer::STATUS_CANCELED :   $transaction->human_status = __( 'Canceled' ); break;
            case StockTransfer::STATUS_DELIVERED :  $transaction->human_status = __( 'Delivered' ); break;
            case StockTransfer::STATUS_DRAFT :      $transaction->human_status = __( 'Draft' ); break;
            case StockTransfer::STATUS_ONGOING :    $transaction->human_status = __( 'Ongoing' ); break;
            case StockTransfer::STATUS_PENDING :    $transaction->human_status = __( 'Pending' ); break;
            case StockTransfer::STATUS_REJECTED :   $transaction->human_status = __( 'Rejected' ); break;
            case StockTransfer::STATUS_STOCKED :    $transaction->human_status = __( 'Stocked' ); break;
        }

        return $this->view( 'NsStockTransfers::transfers.invoice', [
            'title'         =>  __( 'Stock Transfer Invoice' ),
            'description'   =>  __( 'List all products provided within a stock transfer.' ),
            'transaction'   =>  $transaction
        ]);
    }

    public function reviewTransaction( StockTransfer $transaction )
    {
        if ( 
            ( 
                $transaction->status !== StockTransfer::STATUS_REVIEW && $transaction->type === StockTransfer::TYPE_REQUEST
            ) || ( 
                $transaction->status !== StockTransfer::STATUS_PENDING && $transaction->type === StockTransfer::TYPE_TRANSFER
            )
        ) {
            throw new NotAllowedException( __m( 'The transaction is on an invalid status.', 'NsStockTransfers' ) );
        }

        $transaction->load( 'products' );

        return $this->view( 'NsStockTransfers::transfers.review', [
            'title'     =>  sprintf( __m( 'Stock Review : %s', 'NsStockTransfers' ), $transaction->name ),
            'description'   =>  __m( 'Proceed a stock verification for the stock transfer.', 'NsStockTransfers' ),
            'transaction'   =>  $transaction
        ]);
    }

    /**
     * Will proceed the stock transfers and
     * only stock the product that has been counted
     * @param Request $request
     * @param StockTransfer $transaction
     * @return array
     */
    public function confirmTransaction( Request $request, StockTransfer $transaction )
    {
        return $this->stockTransferService->confirmTransaction( $transaction, $request->input( 'counted' ) );
    }

    /**
     * will reject the stock trasnfer
     * @param Request $request
     * @param StockTransfer $transaction
     * @return array
     */
    public function rejectTransaction( Request $request, StockTransfer $transaction )
    {
        return $this->stockTransferService->rejetTransaction( $transaction );
    }
}