<?php
namespace Modules\NsStockTransfers\Services;

use stdClass;
use Exception;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Str;
use App\Exceptions\NotAllowedException;
use App\Exceptions\NotFoundException;
use App\Models\Product;
use App\Models\ProductCategory;
use App\Models\ProductHistory;
use App\Models\Role;
use App\Models\TaxGroup;
use App\Models\Unit;
use App\Models\UnitGroup;
use App\Services\NotificationService;
use App\Services\ProductCategoryService;
use App\Services\ProductService;
use App\Services\TaxService;
use App\Services\UnitService;
use Modules\NsMultiStore\Models\Store;
use Modules\NsStockTransfers\Events\TransactionAfterUpdatedEvent;
use Modules\NsStockTransfers\Events\TransactionAfterApproveEvent;
use Modules\NsStockTransfers\Events\TransactionAfterCanceledEvent;
use Modules\NsStockTransfers\Events\TransactionAfterProcessedEvent;
use Modules\NsStockTransfers\Events\TransactionAfterRejectedEvent;
use Modules\NsStockTransfers\Events\TransactionBeforeApproveEvent;
use Modules\NsStockTransfers\Events\TransactionBeforeCanceledEvent;
use Modules\NsStockTransfers\Events\TransactionBeforeRejectedEvent;
use Modules\NsStockTransfers\Models\StockTransfer;
use Modules\NsStockTransfers\Models\StockTransferProduct;

use Endroid\QrCode\Builder\Builder;
use Endroid\QrCode\Encoding\Encoding;
use Endroid\QrCode\ErrorCorrectionLevel\ErrorCorrectionLevelHigh;
use Endroid\QrCode\Label\Alignment\LabelAlignmentCenter;
use Endroid\QrCode\Label\Font\NotoSans;
use Endroid\QrCode\RoundBlockSizeMode\RoundBlockSizeModeMargin;
use Endroid\QrCode\Writer\PngWriter;
use Illuminate\Support\Collection;

class StockTransferService
{
    /**
     * @var ProductService
     */
    private $productService;

    /**
     * @var UnitService
     */
    private $unitService;

    /**
     * @var ProductCategoryService
     */
    private $categoryService;

    /**
     * @var TaxService
     */
    private $taxService;

    /**
     * @var NotificationService
     */
    private $notificationService;

    public function __construct(
        ProductService $productService,
        UnitService $unitService,
        ProductCategoryService $categorySerivce,
        TaxService $taxService,
        NotificationService $notificationService
    )
    {
        $this->unitService          =   $unitService;
        $this->productService       =   $productService;
        $this->categoryService      =   $categorySerivce;
        $this->taxService           =   $taxService;
        $this->notificationService  =   $notificationService;
    }

    /**
     * Proceed a stock transaction either transfer or request
     * @param array $data
     * @param string $type wether transfer or transaction
     * @return array result of the operation.
     */
    public function updateStockTransaction( $data, StockTransfer $transaction = null ): array
    {
        if ( $data[ 'from_store' ] === $data[ 'to_store' ] ) {
            throw new NotAllowedException( __( 'The source store and the destination store cannot be the same.' ) );
        }

        if ( $data[ 'from_store' ] === Store::current()->id && $data[ 'type' ] === StockTransfer::TYPE_REQUEST && $transaction === null ) {
            throw new NotAllowedException( __( 'A stock request cannot be initialized by a store which is defined as the source.' ) );
        }

        if ( $data[ 'to_store' ] === Store::current()->id && $data[ 'type' ] === StockTransfer::TYPE_TRANSFER ) {
            throw new NotAllowedException( __( 'A stock transfer cannot be initialized by a store which is defined as the destination.' ) );
        }

        if ( $data[ 'status' ] === StockTransfer::STATUS_PENDING && $data[ 'type' ] === StockTransfer::TYPE_TRANSFER ) {
            $this->checkProvidedInventory( $data[ 'products' ] );
        }
    
        $shouldEmit =   false;

        if ( 
            $transaction instanceof StockTransfer &&
            $transaction->status === StockTransfer::STATUS_DRAFT 
            && $data[ 'status' ] === StockTransfer::STATUS_PENDING 
        ) {
            $shouldEmit     =   true;
        } else if (
            $transaction === null &&
            $data[ 'status' ] === StockTransfer::STATUS_PENDING
        ) {
            $shouldEmit     =   true;
        }

        $dateFormat                 =   ns()->date->now()->format( 'y-m-d h:m' );
        $transaction                =   $transaction === null ? new StockTransfer() : $transaction;
        $transaction->status        =   $data[ 'status' ];
        $transaction->name          =   $data[ 'name' ] ?? ( $data[ 'type' ] === StockTransfer::TYPE_REQUEST ? sprintf( __( 'Stock Request %s' ), $dateFormat ) : sprintf( __( 'Stock Transfer %s' ), $dateFormat ) );
        $transaction->type          =   $data[ 'type' ];
        $transaction->author        =   Auth::id();
        $transaction->draft         =   $data[ 'status' ] === 'draft';
        $transaction->from_store    =   $data[ 'from_store' ];
        $transaction->to_store      =   $data[ 'to_store' ];
        $transaction->save();

        foreach( $data[ 'products' ] as $product ) {
            $transactionProduct                         =   isset( $product[ 'id' ] ) ? StockTransferProduct::find( $product[ 'id' ] ) : new StockTransferProduct;
            $transactionProduct->name                   =   $product[ 'name' ];
            $transactionProduct->unit_name              =   $product[ 'unit_quantity' ][ 'unit' ][ 'name' ];
            $transactionProduct->quantity               =   $product[ 'quantity' ];
            $transactionProduct->available_quantity     =   $product[ 'quantity' ];
            $transactionProduct->missing_quantity       =   0;
            $transactionProduct->unit_price             =   $product[ 'unit_quantity' ][ 'sale_price' ];
            $transactionProduct->transfer_id            =   $transaction->id;
            $transactionProduct->product_id             =   $product[ 'product_id' ];
            $transactionProduct->unit_id                =   $product[ 'unit_quantity' ][ 'unit' ][ 'id' ];
            $transactionProduct->total_price            =   ns()->currency->define( $product[ 'unit_quantity' ][ 'sale_price' ] )->multiplyBy( $product[ 'quantity' ] )->getRaw();
            $transactionProduct->barcode                =   rand(1,9) . rand(1,9) . rand(1,9) . rand(1,9) . rand(1,9) . rand(1,9);
            $transactionProduct->save();

            $this->generateQrCode( $transactionProduct );

            /**
             * Assigning a barcode label
             */
            $transactionProduct->barcode        =   str_pad( $transaction->id, 4, '0', STR_PAD_LEFT ) . str_pad( $transactionProduct->id, 4, '0', STR_PAD_LEFT );
            $transactionProduct->save();
        }

        if ( $shouldEmit ) {
            TransactionAfterProcessedEvent::dispatch( $transaction );
        }

        event( new TransactionAfterUpdatedEvent( $transaction ) );

        return [
            'status'    =>  'success',
            'message'   =>  __( 'The stock transfer has been updated correctly' ),
            'data'      =>  compact( 'transaction' )
        ];
    }

    private function generateQrCode( StockTransferProduct $product )
    {
        if ( ! is_dir( storage_path( 'app/public/stock-tranfers' ) ) ) {
            mkdir( storage_path( 'app/public/stock-tranfers' ) );
        }

        $result = Builder::create()
            ->writer(new PngWriter())
            ->writerOptions([])
            ->data( $product->barcode )
            ->encoding(new Encoding('UTF-8'))
            ->errorCorrectionLevel(new ErrorCorrectionLevelHigh())
            ->size(300)
            ->margin(10)
            ->roundBlockSizeMode(new RoundBlockSizeModeMargin())
            ->labelText( $product->name )
            ->labelFont(new NotoSans(20))
            ->labelAlignment(new LabelAlignmentCenter())
            ->build();

        $result->saveToFile( storage_path( 'app/public/stock-tranfers/' . $product->barcode . '.png' ) );

        return [
            'status'    =>  'success',
            'message'   =>  __m( 'The QR Code has been saved.', 'NsStockTranfers' )
        ];
    }

    /**
     * @deprecated
     */
    public function processTransaction( StockTransfer $transaction )
    {
        if ( $transaction->status === StockTransfer::STATUS_APPROVED ) {
            /**
             * let's perform a stock out transaction
             * from the source store
             */
            ns()->store->setStore( Store::find( $transaction->from_store ) );

            /**
             * This should make a stock adjustment removing the 
             * product from the source store.
             */
            if ( $transaction->type === StockTransfer::TYPE_TRANSFER ) {
                $transaction->products->each( function( $product ) {
                    $this->productService->stockAdjustment( ProductHistory::ACTION_TRANSFER_OUT, [
                        'product_id'    =>  $product->product_id,
                        'unit_id'       =>  $product->unit_id,
                        'unit_price'    =>  $product->unit_price,
                        'quantity'      =>  $product->available_quantity
                    ]);
                });
            }

            $transaction->status    =   StockTransfer::STATUS_ONGOING;
            $transaction->save();

        } else if ( $transaction->status === StockTransfer::STATUS_ONGOING ) {
            /**
             * let's perform a stock out transaction
             * from the source store
             */
            ns()->store->setStore( Store::find( $transaction->from_store ) );

            /**
             * This should make a stock adjustment removing the 
             * product from the source store.
             * @deprecated
             */
            if ( $transaction->type === StockTransfer::TYPE_REQUEST ) {
                $transaction->products->each( function( $product ) {
                    $this->productService->stockAdjustment( ProductHistory::ACTION_TRANSFER_OUT, [
                        'product_id'    =>  $product->product_id,
                        'unit_id'       =>  $product->unit_id,
                        'unit_price'    =>  $product->unit_price,
                        'quantity'      =>  $product->available_quantity
                    ]);
                });
            }
        }

        event( new TransactionAfterProcessedEvent( $transaction ) );
    }

    /**
     * Check wether the provided inventory
     * can be processed or throw an error
     * @param $products
     * @return void but can throw Exception
     */
    public function checkProvidedInventory( $products )
    {
        collect( $products )->each( function( $product, $index ) use ( $products ) {
            $totalQuantities   =   collect( $products )->filter( function( $_product, $_index ) use ( $index, $product ) {
                return $_product[ 'product_id' ] === $product[ 'product_id' ] && 
                    $_product[ 'unit_quantity' ][ 'unit' ][ 'id' ] === $product[ 'unit_quantity' ][ 'unit' ][ 'id' ] && 
                    $index !== $_index;
            })->map( function( $product ) {
                return $product[ 'quantity' ];
            })->sum();

            $unitQuantity   =   $this->productService->getUnitQuantity( $product[ 'product_id' ], $product[ 'unit_quantity' ][ 'unit' ][ 'id' ] );

            if ( $unitQuantity->quantity < $totalQuantities + $product[ 'quantity' ] ) {
                throw new Exception( sprintf( __( 'Not enough stock for the product %s using the unit %s'), $product[ 'name' ], $product[ 'unit_quantity' ][ 'unit' ][ 'name' ] ) );
            }
        });
    }

    /**
     * Perform a stock transaction rejection
     * @param StockTransfer $transaction
     * @return array
     */
    public function rejetTransaction( StockTransfer $transaction )
    {
        event( new TransactionBeforeRejectedEvent( $transaction ) );

        // $this->performStockReturn( $transaction, ProductHistory::ACTION_TRANSFER_REJECTED );

        $transaction->status    =   StockTransfer::STATUS_REJECTED;
        $transaction->save();

        event( new TransactionAfterRejectedEvent( $transaction ) );

        return [
            'status'    =>  'success',
            'message'   =>  __( 'The stock transaction has been correctly rejected' )
        ];
    }

    private function performStockReturn( StockTransfer $transaction, $action )
    {
        /**
         * This should make a stock adjustment removing the 
         * product from the source store.
         */
        $store      =   Store::find( $transaction->from_store );
        
        /**
         * let's define the store
         * before performing the rejection
         */
        ns()->store->setStore( $store );

        $transaction->products->each( function( $product ) use  ( $action ) {
            $this->productService->stockAdjustment( $action, [
                'product_id'    =>  $product->product_id,
                'unit_id'       =>  $product->unit_id,
                'unit_price'    =>  $product->unit_price,
                'quantity'      =>  $product->available_quantity
            ]);
        });
    }

    /**
     * Perform a stock transaction cancelation
     * @param StockTransfer $transaction
     * @return array
     */
    public function cancelTransaction( StockTransfer $transaction )
    {
        event( new TransactionBeforeCanceledEvent( $transaction ) );
        
        /**
         * only when the transaction is a transfer that is about
         * to be canceled, we can perform a stock return.
         */
        if ( $transaction->type === StockTransfer::TYPE_TRANSFER ) {
            // $this->performStockReturn( $transaction, ProductHistory::ACTION_TRANSFER_CANCELED );
        }

        $transaction->status    =   StockTransfer::STATUS_CANCELED;
        $transaction->save();

        event( new TransactionAfterCanceledEvent( $transaction ) );

        return [
            'status'    =>  'success',
            'message'   =>  __( 'The stock transaction has been correctly canceled.' )
        ];
    }

    /**
     * Will approve the provider transaction
     * @param StockTransfer $transaction
     * @param boolean $force
     * @return array $response
     */
    public function approveTransaction( StockTransfer $transaction, $force = false )
    {
        /**
         * let's retreive the products
         * from the source store
         */
        $store      =   Store::find( $transaction->from_store );

        /**
         * let's define the store
         * before for a nice table prefixing
         */
        ns()->store->setStore( $store );

        /**
         * Make sure the transaction is loaded
         * along with his products
         */
        $transaction->load( 
            'products.product.category', 
            'products.unit.group', 
            'products.product.tax_group' 
        );
        
        event( new TransactionBeforeApproveEvent( $transaction ) );

        /**
         * This should make a stock adjustment removing the 
         * product from the source store.
         */
        $destinationStore       =   Store::find( $transaction->to_store );
        $fromStore              =   Store::find( $transaction->from_store );

        /**
         * This will perform a verification to check
         * if the stock that is about to be transfered actually
         * exists on the system.
         */
        $references   =   $this->checkDataExistence( $transaction, $destinationStore );

        if ( ns()->rootOption->get( 'ns-stocktransfers-approval', 'no' ) === 'no' || $force === true ) {

            /**
             * let's perform a stock out transaction
             * from the source store
             */
            Store::switchTo( $fromStore );

            /**
             * This should make a stock adjustment removing the 
             * product from the source store.
             */
            $transaction->products->each( function( $product ) {
                $this->productService->stockAdjustment( ProductHistory::ACTION_TRANSFER_OUT, [
                    'product_id'    =>  $product->product_id,
                    'unit_id'       =>  $product->unit_id,
                    'unit_price'    =>  $product->unit_price,
                    'quantity'      =>  $product->available_quantity
                ]);
            });

            /**
             * let's define the store
             * before for a nice table prefixing
             */
            Store::switchTo( $destinationStore );

            $references->each( function( $reference ) {
                $this->productService->stockAdjustment( ProductHistory::ACTION_TRANSFER_IN, [
                    'product_id'    =>  $reference->product->id, // the product that has been created
                    'unit_id'       =>  $reference->unit->id, // the unit that has been created or that was arleady existing
                    'unit_price'    =>  $reference->stockTransferProduct->unit_price, // unit price as defined on the stockTransferProduct
                    'quantity'      =>  $reference->stockTransferProduct->available_quantity // the quantity as defined on the stockTransferProduct
                ]);
            });
            
            $transaction->status    =   StockTransfer::STATUS_STOCKED;
        } else {

            /**
             * On approval, the stock is not deducted. 
             * It will be deducted when the transfered
             */
            $transaction->status    =   StockTransfer::STATUS_REVIEW;
        }

        $transaction->save();

        event( new TransactionAfterApproveEvent( $transaction ) );

        return [
            'status'    =>  'success',
            'message'   =>  __( 'The stock transfer has been correctly approved.' ),
        ];
    }

    /**
     * Will perform a verification over the existing transaction
     * @param StockTransfer $transaction
     * @param Store $destinationStore
     * @return Collection
     */
    private function checkDataExistence( StockTransfer $transaction, Store $destinationStore )
    {
        return $transaction->products->map( function( StockTransferProduct $stockTransferProduct ) use ( $destinationStore ) {
            
            $product    =   $stockTransferProduct->product;

            /**
             * Switch to the destination store
             * to check if the product/unit with the same
             * SKU exists there
             */
            Store::switchTo( $destinationStore );

            if ( ! $product instanceof Product ) {
                throw new NotFoundException( sprintf( __( 'The product attached to the stock transfer product named "%s" doesn\'t have any reference.' ), $stockTransferProduct->name ) );
            }
            
            if ( ! $stockTransferProduct->unit instanceof Unit ) {
                throw new NotFoundException( sprintf( __( 'The unit attached to the product "%s" cannot be found.' ), $stockTransferProduct->name ) );
            }

            if ( ! $product->category instanceof ProductCategory ) {
                throw new NotFoundException( sprintf( __( 'The category attached to the product "%s" cannot be found.' ), $stockTransferProduct->name ) );
            }

            /**
             * we'll check wether the attached
             * resources exists on the destination
             * store.
             */
            if ( $product->tax_group instanceof TaxGroup ) {
                $taxGroup        =   $this->taxService->getTaxGroupUsingName( $product->tax_group->name );
                if ( ! $taxGroup instanceof TaxGroup ) {
                    $taxGroup   =   $this->createTaxGroup( $product->tax_group );
                }
            }

            $unit        =   $this->unitService->getUsingIdentifier( $stockTransferProduct->unit->identifier );
            if ( ! $unit instanceof Unit ) {
                $unit   =   $this->createUnit( $stockTransferProduct->unit, $destinationStore );
            }

            $category    =   $this->categoryService->getUsingName( $product->category->name );
            if ( ! $category instanceof ProductCategory ) {
                $category   =   $this->createCategory( $product->category );
            }
            
            /**
             * the product doesn't exists we should create
             * a new reference just by copying the source
             * store some product details.
             */
            $newProduct         =   $this->productService->getProductUsingSKU( $product->sku );
            if ( ! $newProduct instanceof Product ) {
                $newProduct     =   $this->createProduct( $product, $unit, $category, $stockTransferProduct );
            }

            $finalProduct                           =   new stdClass;
            $finalProduct->unit                     =   $unit;
            $finalProduct->category                 =   $category;

            if ( isset( $taxGroup ) ) {
                $finalProduct->taxGroup             =   $taxGroup;
            }
            
            $finalProduct->stockTransferProduct     =   $stockTransferProduct;
            $finalProduct->product                  =   $newProduct;

            return $finalProduct;
        });
    }

    /**
     * create a specific unit and assign it to a unit group
     * @param Unit $unit
     * @return Unit
     */
    private function createUnit( Unit $unit )
    {
        $group  =   UnitGroup::where( 'name', $unit->group->name );

        if ( ! $group instanceof UnitGroup ) {
            $response   =   $this->unitService->createGroup([
                'name'          =>  $unit->group->name,
                'description'   =>  $unit->group->description
            ]);

            $group      =   $response[ 'data' ][ 'group' ];
        }
        
        $response   =   $this->unitService->createUnit([
            'group_id'      =>  $group->id,
            'name'          =>  $unit->name,
            'identifier'    =>  $unit->identifier,
            'value'         =>  $unit->value,
            'base_unit'     =>  $unit->base_unit,
            'preview_url'   =>  $unit->preview_url,
        ]);

        return $response[ 'data' ][ 'unit' ];
    }

    /**
     * create a single category on the
     * actual store using the provided ProductCategory
     * @param ProductCategory $category
     * @return ProductCategory created on the actual store
     */
    private function createCategory( ProductCategory $category )
    {
        $response   =   $this->categoryService->create([
            'name'  =>  $category->name,
            'description'   =>  $category->description,
            'parent_id'     =>  $category->parent_id,
        ]);

        return $response[ 'data' ][ 'category' ];
    }

    private function createProduct( Product $product, Unit $unit, ProductCategory $category, StockTransferProduct $stockTransferProduct )
    {
        $data           =   $product->toArray();
        $finalProduct   =   [];

        foreach( $data as $key => $value ) {
            if ( ! in_array( $key, [
                'id',
                'author',
                'created_at',
                'updated_at',
                'category_id',
                'tax_group',
            ]) ) {
                $finalProduct[ $key ]   =   $value;
            }
        }

        /**
         * @todo we might need to upload images to destination store
         * we'll figure out if that's necessary with time.
         * @todo transfer might need to support wholesale_price as well
         */
        $finalProduct[ 'images' ]       =   []; // should we copy the same images url from the product or create new one ?
        $finalProduct[ 'category_id' ]  =   $category->id;
        $finalProduct[ 'unit_group' ]   =   $unit->group->id;
        $finalProduct[ 'units' ]        =   [
            'unit_group'                =>  $unit->group->id,
            'selling_group'             =>  [
                [
                    'unit_id'               =>  $unit->id,
                    'sale_price_edit'       =>  $stockTransferProduct->unit_price,
                    'wholesale_price_edit'  =>  0,
                    'preview_url'           =>  '', // should we copy the same URL or import the product ?
                ]
            ]
        ];

        $response   =   $this->productService->create( $finalProduct );
        return $response[ 'data' ][ 'product' ];
    }

    private function createTaxGroup( TaxGroup $group )
    {
        $group->load( 'taxes' );

        $response           =   $this->taxService->createTaxGroup([
            'name'          =>  $group->name,
            'description'   =>  $group->description,
        ]);

        $newGroup   =   $response[ 'data' ][ 'group' ];

        $group->taxes->map( function( $tax ) use( $newGroup ) {
            $tax->tax_group_id      =   $newGroup->id;
            return $this->taxService->createTax( $tax->toArray() );
        });

        return $newGroup;
    }

    public function notify( StockTransfer $transfer )
    {
        if ( $transfer->type === StockTransfer::TYPE_REQUEST && $transfer->status === StockTransfer::STATUS_STOCKED ) {

            Store::switchTo( $transfer->to_store );

            $this->notificationService->create([
                'title'         =>      __( 'Stock Request Approved' ),
                'description'   =>      sprintf(
                    __( 'The stock request "%s" has been aproved and stocked.' ),
                    $transfer->name
                ),
                'identifier'    =>      'stocktransfer-approval-' . $transfer->id,
                'url'           =>      ns()->route( 'ns.stocktransfer-invoice', [
                    'transaction'   =>  $transfer->id
                ])
            ])->dispatchForGroup([
                Role::namespace( 'admin' ),
                Role::namespace( 'nexopos.store.administrator' ),
            ]);
        }

        if ( $transfer->type === StockTransfer::TYPE_REQUEST && $transfer->status === StockTransfer::STATUS_ONGOING ) {
            
            Store::switchTo( $transfer->from_store );

            $this->notificationService->create([
                'title'         =>      __( 'Stock Request Awaiting' ),
                'description'   =>      sprintf(
                    __( 'A new stock request "%s" is awaiting approval.' ),
                    $transfer->name
                ),
                'identifier'    =>      'stocktransfer-awaiting-' . $transfer->id,
                'url'           =>      ns()->route( 'ns.stocktransfer-invoice', [
                    'transaction'   =>  $transfer->id
                ])
            ])->dispatchForGroup([
                Role::namespace( 'admin' ),
                Role::namespace( 'nexopos.store.administrator' ),
            ]);
        }

        if ( $transfer->type === StockTransfer::TYPE_TRANSFER && $transfer->status === StockTransfer::STATUS_PENDING ) {
            Store::switchTo( $transfer->to_store );

            $this->notificationService->create([
                'title'         =>      __( 'Stock Transfert Awaiting' ),
                'description'   =>      sprintf( 
                    __( 'A new stock transfer "%s" is awaiting stocking confirmation.' ),
                    $transfer->name
                ),
                'identifier'    =>      'stocktransfer-awaiting-' . $transfer->id,
                'url'           =>      ns()->route( 'ns.stocktransfer-invoice', [
                    'transaction'   =>  $transfer->id
                ])
            ])->dispatchForGroup([
                Role::namespace( 'admin' ),
                Role::namespace( 'nexopos.store.administrator' ),
            ]);
        }

        if ( $transfer->type === StockTransfer::TYPE_TRANSFER && $transfer->status === StockTransfer::STATUS_STOCKED ) {

            Store::switchTo( $transfer->from_store );

            $this->notificationService->create([
                'title'         =>      __( 'Stock Transfert Stocked' ),
                'description'   =>      sprintf( 
                    __( 'The stock transfer "%s" has been approved and stocked by the destination store.' ),
                    $transfer->name
                ),
                'identifier'    =>      'stocktransfer-awaiting-' . $transfer->id,
                'url'           =>      ns()->route( 'ns.stocktransfer-invoice', [
                    'transaction'   =>  $transfer->id
                ])
            ])->dispatchForGroup([
                Role::namespace( 'admin' ),
                Role::namespace( 'nexopos.store.administrator' ),
            ]);
        }

        if ( $transfer->type === StockTransfer::TYPE_TRANSFER && $transfer->status === StockTransfer::STATUS_REJECTED ) {

            Store::switchTo( $transfer->from_store );

            $this->notificationService->create([
                'title'         =>      __( 'Stock Transfert Rejected' ),
                'description'   =>      sprintf( 
                    __( 'The stock transfer "%s" has been rejected by the destination store.' ),
                    $transfer->name
                ),
                'identifier'    =>      'stocktransfer-rejected-' . $transfer->id,
                'url'           =>      ns()->route( 'ns.stocktransfer-invoice', [
                    'transaction'   =>  $transfer->id
                ])
            ])->dispatchForGroup([
                Role::namespace( 'admin' ),
                Role::namespace( 'nexopos.store.administrator' ),
            ]);
        }

        if ( $transfer->type === StockTransfer::TYPE_REQUEST && $transfer->status === StockTransfer::STATUS_REJECTED ) {

            Store::switchTo( $transfer->to_store );

            $this->notificationService->create([
                'title'         =>      __( 'Stock Request Rejected' ),
                'description'   =>      sprintf( 
                    __( 'The stock request "%s" has been rejected by the source store.' ),
                    $transfer->name
                ),
                'identifier'    =>      'stocktransfer-rejected-' . $transfer->id,
                'url'           =>      ns()->route( 'ns.stocktransfer-invoice', [
                    'transaction'   =>  $transfer->id
                ])
            ])->dispatchForGroup([
                Role::namespace( 'admin' ),
                Role::namespace( 'nexopos.store.administrator' ),
            ]);
        }
    }

    public function confirmTransaction( StockTransfer $transaction, $counted )
    {
        foreach( $counted as $product ) {
            $productRef     =   StockTransferProduct::where( 'barcode', $product[ 'barcode' ] )
                ->first();

            if ( $productRef instanceof StockTransferProduct ) {
                $productRef->available_quantity     =   $product[ 'available_quantity' ];
                $productRef->missing_quantity       =   $product[ 'missing_quantity' ];
                $productRef->approval_status        =   StockTransferProduct::APPROVAL_ACCEPTED;
                $productRef->quantity               =   $product[ 'available_quantity' ];
                $productRef->save();
            }
        }

        return $this->approveTransaction( $transaction, true );
    }
}